<?php
/**
 * Copyright (C) 2004-2021 "Setevye informacionnye sistemy", Ltd (Russia)
 *                          en: "Network Information Systems"
 */

require 'helpers.php';
require (__DIR__ . '/../memcache.php');
require (__DIR__ . '/../object2array.php');

class OnvifClientException extends Exception
{
}

class OnvifServices
{
    const DEVICEMANAGEMENT = 'devicemgmt';
    const MEDIA = 'media';
    const PTZ = 'ptz';
}

class OnvifClient
{
    const REQUEST_TIME_AVG = 1;

    /**
     * @var SoapClient
     */
    private $soap;

    private $cache;

    /**
     * Instances of SoapClient for each service.
     * @var array
     */
    private $soapClients = array();

    private $origin;
    private $baseLocation;
    private $cacheKey; // mp5 hash of origin
    private $cacheData;

    private $deltaTime; // client and server timeoffset in seconds for auth

    private $credentials = array();
    private $headers = array();

    private $logSoapRequests = false;
    private $loggedSoapRequests = array();

    private $lastRequestService;
    private $lastRequestTime;
    private $lastRequestMethodName;
    private $lastRequestSuccess;

    protected $capabilities = array();
    protected $allowServices = array(); // [$serviceName][$onvifSpecVer] = [$relURI,$ver[,$caps]]


    // флаг для др. функций, чтобы было понятно что они вызваны из-вне, а не из конструктора
    private $not_in_contructor = false;

    private $defaultOptions = array(
        'proto' => 'http',
        'host' => null,
        'port' => null,
        'username' => null,
        'password' => '',
        'logSoapRequests' => false,
        'wsdlLocation' => __DIR__ . '/wsdl',
        'readTimeout' => 10
    );

    public function buildBaseLocation(array $options)
    {
        if ($options['proto'] !== 'http' && $options['proto'] !== 'https') {
            throw new OnvifClientException('Unexpected "proto" option, must be an "http" or "https", found "' . $options['proto'] . '"');
        }

        if (empty($options['host'])) {
            throw new OnvifClientException('"host" value must not to be empty');
        }

        $origin = $options['host'];
        if (!is_null($options['port'])) {
            if ($options['proto'] === 'https') {
                if ($options['port'] != 443) {
                    $origin .= ':' . $options['port'];
                }
            } else {
                if ($options['port'] != 80) {
                    $origin .= ':' . $options['port'];
                }
            }
        }

        $baseLocation = $options['proto'] . '://' .  $origin;
        if (filter_var($baseLocation, FILTER_VALIDATE_URL) === false) {
            throw new OnvifClientException('Not a valid baseLocation, must be an URI, found : ' . $baseLocation);
        }

        $this->origin = $origin;
        $this->baseLocation = $baseLocation;
    }

    /**
     * Options:
     * @param array $options
     * @throws OnvifClientException
     */
    public function __construct(array $options = array())
    {
        $options = array_merge($this->defaultOptions, $options);

        // setup origin and baseLocation properties
        $this->buildBaseLocation($options);

        if (!empty($options['username'])) {
            $this->setCredentials($options['username'], $options['password']);
        }

        if ($options['wsdlLocation'] != $this->defaultOptions['wsdlLocation']
            && filter_var($options['wsdlLocation'], FILTER_VALIDATE_URL) === false) {
            throw new OnvifClientException('Not a valid wsdlLocation, must be an URI, found  : ' . $options['wsdlLocation']);
        }
        $this->wsdlLocation = $options['wsdlLocation'];

        $this->logSoapRequests = $options['logSoapRequests'];

        $this->readTimeout = max((int)$options['readTimeout'], 3);

        $this->soap = $this->getSoapClientForService(OnvifServices::DEVICEMANAGEMENT);

        $this->cacheKey = hash('md5', $this->origin);
        $this->cache = new \Avreg\Cache(array('debug' => false));
        $this->cacheData = $this->cache->lock($this->cacheKey);

        $this->deltaTime = array_key_exists('deltaTime', $this->cacheData) ? $this->cacheData['deltaTime'] : false /* as first time */;
        if ($this->deltaTime === false) {
            $this->cacheData['deltaTime'] = $this->timeSync();
            $caps = $services = null;
            $r = $this->doSoapRequest(
                \OnvifServices::DEVICEMANAGEMENT, 'GetServices', array('IncludeCapability' => true));
            if ($r['isOk']) {
                // Axis v5.? not support GetServices()
                $services = $r['result'];
            } else {
                $r = $this->doSoapRequest(\OnvifServices::DEVICEMANAGEMENT, 'GetCapabilities');
                if ($r['isOk']) {
                    $caps = $r['result'];
                }
            }

            $this->parseCapsAndServices($caps, $services);
            if (empty($this->capabilities) && empty($this->allowServices)) {
                $this->resetSession(true);
                throw new OnvifClientException('Could not load ONVIF capabilities and services both');
            }
            $this->cacheData['allowServices'] = $this->allowServices;
            $this->cacheData['capabilities'] = $this->capabilities;
            /*
                error_log('+++++++++++++++++++++ allowServices ++++++++++++++++++++');
                error_log(print_r($this->allowServices, true));
                error_log('+++++++++++++++++++++ capabilities ++++++++++++++++++++');
                error_log(print_r($this->capabilities, true));
            */
        } else {
            if (!@empty($this->cacheData['allowServices'])) {
                $this->allowServices = $this->cacheData['allowServices'];
            }

            if (!@empty($this->cacheData['capabilities'])) {
                $this->capabilities = $this->cacheData['capabilities'];
            }
        }
        // error_log($this->cacheKey . ' deltaTime = [' .  $this->deltaTime . ']');
        $this->cache->unlock($this->cacheKey, $this->cacheData);

        // флаг для др. функций, чтобы было понятно что они вызваны из-вне, а не из конструктора
        $this->not_in_contructor = true;
    }

    private function resetSession($now_locked = false)
    {
        $this->deltaTime = false;
        $this->allowServices = array();
        $this->capabilities = array();

        $this->cacheData = array(
            'deltaTime' => false,
            'allowServices' => array(),
            'capabilities' => array()
        );
        if ($now_locked) {
            $this->cache->unlock($this->cacheKey, $this->cacheData);
        } else {
            $this->cache->set($this->cacheKey, $this->cacheData);
        }
    }

    protected function &getServiceConf($service, $onvifSpecVer = null)
    {
        if (!$onvifSpecVer) {
            $allSpecVers = array_keys($this->allowServices[$service]);
            ksort($allSpecVers);
            for ($i = 0; $i < count($allSpecVers); $i += 1) {
                $spec = $allSpecVers[$i];
                $conf =&$this->allowServices[$service][$spec];
                if (!empty($conf['ver'])) {
                    // by GetServices()
                    $onvifSpecVer = $spec;
                    break;
                }
            }
            if (!$onvifSpecVer) {
                $onvifSpecVer = $allSpecVers[0];
            }
        }
        if (!$onvifSpecVer) {
            $onvifSpecVer = 'ver10';
        }

        return $this->allowServices[$service][$onvifSpecVer];
    }

    protected function getSoapClientForService($service, $onvifSpecVer = null)
    {
        // FIXME: https://stackoverflow.com/questions/10652705/error-fetching-http-body-with-php-soapclient
        if (!isset($this->soapClients[$service])) {
            $http_opt = array(
                'timeout' => $this->readTimeout
            );

            if ($service === \OnvifServices::DEVICEMANAGEMENT) {
                $servicePath = '/onvif/device_service';
                $onvifSpecVer = 'ver10';
            } else {
                if (isset($this->allowServices[$service])) {
                    $serviceConf = $this->getServiceConf($service, $onvifSpecVer);
                    $onvifSpecVer = $serviceConf['onvifSpecVer'];
                    $servicePath = $serviceConf['uri'];
                } else {
                    throw new OnvifClientException('Unexpected onvif service "' . $service . '"');
                }
            }

            $options = array(
                'location' => $this->baseLocation . $servicePath,
                'soap_version' => SOAP_1_2,
                'login' => $this->credentials['username'],
                'password' => $this->credentials['password'],
                'authentication' =>  SOAP_AUTHENTICATION_DIGEST,
                'stream_context' => stream_context_create(array('http' => $http_opt)), // FIXME why only http
                'cache_wsdl' => WSDL_CACHE_MEMORY,
                'keep_alive' => true,
                'compression' => SOAP_COMPRESSION_ACCEPT | SOAP_COMPRESSION_GZIP,
                'trace' => 1
            );

            $this->soapClients[$service] = new SoapClient(
                $this->getWsdlForService($service, $onvifSpecVer),
                $options
            );
        }

        return $this->soapClients[$service];
    }

    protected function getWsdlForService($service, $onvifSpecVer)
    {
        return $this->wsdlLocation . '/' . $onvifSpecVer . '/' . $service . '.wsdl';
    }

    protected function setCredentials(string $username, string $password = '')
    {
        $this->credentials = array(
            'username' => $username,
            'password' => $password,
        );
    }

    public function hasCredentials()
    {
        return empty($this->credentials);
    }

    protected function timeSync()
    {
        // store delta from server time, need for generating security nonce
        $response = $this->doSoapRequest(
            OnvifServices::DEVICEMANAGEMENT,
            'GetSystemDateAndTime',
            array(),
            false
        );

        if ($response['isOk'] === true) {
            $response = object2array($response['result']);

            $deviceDateTime = $response['SystemDateAndTime']['UTCDateTime'];

            $deviceTimestamp = gmmktime(
                $deviceDateTime['Time']['Hour'],
                $deviceDateTime['Time']['Minute'],
                $deviceDateTime['Time']['Second'],
                $deviceDateTime['Date']['Month'],
                $deviceDateTime['Date']['Day'],
                $deviceDateTime['Date']['Year']
            );

            $this->deltaTime = time() - $deviceTimestamp - self::REQUEST_TIME_AVG;
        } else {
            $this->deltaTime = false;
        }
        return $this->deltaTime;
    } /* timeSync() */

    private function passwordDigest($username, $password, $utcTime = "default", $nonce = "default")
    {
        if ($utcTime == 'default') {
            $utcTime = gmdate('Y-m-d\TH:i:s.000\Z');
        }
        if ($nonce == 'default') {
            $nonce = mt_rand();
        }
        $digest = array();
        $passdigest = base64_encode(
            pack('H*', sha1(pack('H*', $nonce) . pack('a*', $utcTime) . pack('a*', $password)))
        );
        //$passdigest=base64_encode(sha1($nonce.$utcTime.$password,true)); // alternative
        $digest['username'] = $username;
        $digest['passwordDigest'] = $passdigest;
        $digest['nonce'] = base64_encode(pack('H*', $nonce));
        //$digest['nonce'] = base64_encode($nonce); // alternative
        $digest['created'] = $utcTime;
        return $digest;
    }

    private function setAuthHeaders()
    {
        if (isset($this->credentials['username']) && isset($this->credentials['password'])) {
            $token = $this->passwordDigest(
                $this->credentials['username'],
                $this->credentials['password'],
                gmdate('Y-m-d\TH:i:s.000\Z', time() - $this->deltaTime));

            $ns = 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd';

            /* ------------------------------------------------------------------------------------------------------ */
            // more proper way of creating
            // need adjustment

            /*
            $objSoapVarWSSEAuth = new SoapVar(
                array(
                    'Username' => new SoapVar($this->credentials['username'], XSD_STRING, null, null, null, $ns),
                    'Password' => new SoapVar($token['passwordDigest'], XSD_STRING, 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#PasswordDigest', null, null, $ns),
                    'Nonce' => new SoapVar($token['nonce'], XSD_STRING, null, $ns, null, $ns),
                    'Created' => new SoapVar($token['created'], XSD_STRING, null, null, null, "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd"),
                ),
                SOAP_ENC_OBJECT,  null, $ns, null, $ns);

            $objSoapVarWSSEToken = new SoapVar(
                array('UsernameToken' => $objSoapVarWSSEAuth), SOAP_ENC_OBJECT,  null, $ns, null, $ns
            );

            $objSoapVarWSSEHeader = new SoapHeader(null, 'Security', $objSoapVarWSSEToken, true);

            // todo - find out why can't set headers as array
            $this->headers = $objSoapVarWSSEHeader;*/

            /* ------------------------------------------------------------------------------------------------------ */

            $xml = '
                <wsse:Security env:mustUnderstand="1" xmlns:wsse="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd">
                    <wsse:UsernameToken>
                        <wsse:Username>' . $token['username'] . '</wsse:Username>
                        <wsse:Password Type="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#PasswordDigest">' . $token['passwordDigest'] . '</wsse:Password>
                        <wsse:Nonce EncodingType="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary">' . $token['nonce'] . '</wsse:Nonce>
                        <wsu:Created xmlns:wsu="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd">' . $token['created'] . '</wsu:Created>
                    </wsse:UsernameToken>
                </wsse:Security>
            ';

            $this->headers = new SoapHeader($ns, 'Security', new SoapVar($xml, XSD_ANYXML), true);
        } else {
            throw new OnvifClientException('User credentials are incorrect.');
        }
    }

    public function doSoapRequest($service, $action, $params = array(), $auth = true)
    {
        /*
        error_log(
            sprintf(
                'doSoapRequest() service=%s action=%s auth=%s deltaTime=%s',
                $service, $action, (string)$auth, (string)$this->deltaTime
            )
        );
        */

        try {
            if ($auth) {
                $this->setAuthHeaders();
            }

            $this->lastRequestService = $service;
            $this->lastRequestTime = time();
            $this->lastRequestMethodName = $action;

            $client = $this->getSoapClientForService($service);

            $client->__setSoapHeaders($this->headers);
            $result = call_user_func(array($client, $action), $params);

            $this->lastRequestSuccess = true;

            if ($this->logSoapRequests) {
                $this->loggedSoapRequests[] = $this->soapDebug();
            }

            // flush headers
            $this->headers = array();

            return array(
                'isOk' => true,
                'result' => $result,
                'debug' => $this->soapDebug()
            );
        } catch (Exception $e) {
            /// XXX to do reset session on some case
            if ($e instanceof SoapFault or $e instanceof OnvifClientException) {
                $errMsg = $e->getMessage();
                error_log($this->origin . ': onvif ' . $service . '->' . $action . '() failed with: ' . $errMsg);
            }

            if ($e instanceof SoapFault) {
                if ($this->not_in_contructor and (stristr($errMsg, 'auth') or stristr($errMsg, 'credential'))) {
                    $this->resetSession();
                }
            }

            if ($e instanceof SoapFault or $e instanceof OnvifClientException) {
                if ($this->logSoapRequests) {
                    $this->lastRequestSuccess = false;
                    $this->loggedSoapRequests[] = $this->soapDebug();
                }
                return array(
                    'isOk' => false,
                    'exception' => $e
                );
            } else {
                // Keep throwing it.
                throw $e;
            }
        }
    }

    public function getLoggedSoapRequests()
    {
        return $this->loggedSoapRequests;
    }

    /**
     * @return array
     */
    public function soapDebug()
    {
        $client = $this->getSoapClientForService($this->lastRequestService);

        return array(
            'serviceName' => $this->lastRequestService,
            'requestSuccess' => $this->lastRequestSuccess,
            'requestTime' => date("Y-m-d H:i:s", $this->lastRequestTime),
            'methodName' => $this->lastRequestMethodName,
            'requestHeaders' => $client->__getLastRequestHeaders(),
            'request' => $client->__getLastRequest(),
            'responseHeaders' => $client->__getLastResponseHeaders(),
            'response' => $client->__getLastResponse()
        );
    }

    public function soapDebugHtmlOutput()
    {
        $debug = $this->soapDebug();

        return $output = '<code>' . nl2br(htmlspecialchars($debug['requestHeaders'], true)) . '</code>'
            . highlight_string(xmlpp($debug['request']), true) . "<br/>\n"
            . '<code>' . nl2br(htmlspecialchars($debug['responseHeaders'], true)) . '</code>' . "<br/>\n"
            . highlight_string(xmlpp($debug['response']), true) . "<br/>\n";
    }


    protected function normalizeServiceName ($name)
    {
        $upname = strtoupper($name);
        if (0 === strpos($name, $upname)) {
            return strtolower($name);
        } else {
            return lcfirst($name);
        }
    }

    protected function pullService(
        $name,
        $uri,
        $onvifSpecVer = null,
        $version = null,
        $capabilities = null
    )
    {
        if (empty($name) || empty($uri)) {
            throw new OnvifClientException('name|uri values must not to be empty');
        }

        $name = $this->normalizeServiceName($name);

        if (empty($onvifSpecVer)) {
            if (in_array($name, ['imaging', 'ptz', 'analytics'])) {
                $onvifSpecVer = 'ver20';
             } else {
                $onvifSpecVer = 'ver10';
             }
        }


        if (!is_readable($this->wsdlLocation . '/' . $onvifSpecVer . '/' . $name . '.wsdl')) {
            error_log('Skip unsupported "' . $onvifSpecVer . '/' . $name . '" service');
            return false;
        }

        if (!isset($this->allowServices[$name])) {
            $this->allowServices[$name] = array();
        }

        if (array_key_exists($onvifSpecVer, $this->allowServices[$name]) && $version /* GetCapabilities() has no service version */)
        {
            // has already, might second GetServices()
            // or GetCapabilities after GetServices() from contructor()
            if ($capabilities && empty($this->allowServices[$name][$onvifSpecVer].caps)) {
                $this->allowServices[$name][$onvifSpecVer]['caps'] = $capabilities;
            }
        } else {
            $this->allowServices[$name][$onvifSpecVer] = array(
                'uri' => self::getRelativeURI($uri),
                'ver' => $version,
                'onvifSpecVer' => $onvifSpecVer,
                'caps' => $capabilities
            );
        }

        return true;
    }

    protected static function getRelativeURI($fullUri)
    {
        $u = parse_url($fullUri);
        $rel = @empty($u['path']) ? '/' : $u['path'];
        if (!@empty($u['query'])) {
            $rel .= '?' . $u['query'];
        }
        if (!@empty($u['fragment'])) {
            $rel .= '#' . $u['fragment'];
        }
        return $rel;
    }

    protected function parseCapsAndServices($allCapsObj = null, $allServObj = null)
    {
        if (!is_null($allServObj)) {
            foreach ($allServObj->Service as &$servObj) {
                $namespacePath = parse_url($servObj->Namespace, PHP_URL_PATH);
                $version = (array)$servObj->Version;
                if (preg_match ('/\/ver(\d+)\/(\w+)\/wsdl$/', $namespacePath, $matches)) {
                    $onvifSpecVer = 'ver' . $matches[1];
                    $serviceName = $matches[2];
                    $this->pullService($serviceName, $servObj->XAddr, $onvifSpecVer, $version);
                } else {
                    error_log('Unexpected onvif service namespace "' . $servObj->Namespace . '"');
                }
            }
        }
        if (!is_null($allCapsObj)) {
            $this->capabilities = object2array($allCapsObj->Capabilities, array('XAddr'));
            if (is_null($allServObj)) {
                // придётся самому строить services из capabilities
                foreach ($allCapsObj->Capabilities as $capName => &$capValObj) {
                    if (property_exists($capValObj, 'XAddr')) {
                        $this->pullService($capName, $capValObj->XAddr);
                    }
                }
                if (property_exists($allCapsObj->Capabilities, 'Extension')) {
                    foreach ($allCapsObj->Capabilities->Extension as $capName => &$capValObj) {
                        if (property_exists($capValObj, 'XAddr')) {
                            $this->pullService($capName, $capValObj->XAddr);
                        }
                    }
                }
            }
        }
    }
}
