<?php
/**
 * Copyright (C) 2004-2021 "Setevye informacionnye sistemy", Ltd (Russia)
 *                          en: "Network Information Systems"
 */

namespace Avreg;

require_once(__DIR__ . '/../../head-xhr.inc.php');
require_once(__DIR__ . '/../OnvifClient/OnvifAjaxController.php');
require_once(__DIR__ . '/../AjaxController.php');
require_once(__DIR__ . '/PtzInterface.inc.php');
require_once(__DIR__ . '/../../lib/get_cams_params.inc.php');

/***
 * class OnvifPtzController
 *
 * OnvifPtzController -> OnvifAjaxController -> AjaxController
 * OnvifAjaxController->connect(
 *    $this->onvifClient = new \OnvifClient()
 * )
 *
 */
class OnvifPtzController extends OnvifAjaxController implements PtzInterface
{
    protected function getConnectionData($requestData = array())
    {
        global $adb;

        if (!isset($requestData['cameraNumber']) || !settype($requestData['cameraNumber'], 'integer')) {
            throw new \Exception('cameraNumber not set or invalid');
        }
        $camNr = &$requestData['cameraNumber'];

        $cams_params = get_cams_params(array(
            'work',
            'InetCam_IP',
            'InetCam_http_port',
            'text_left',
            'InetCam_USER',
            'InetCam_PASSWD',
            'onvif_profile_token',
            'onvif_http_port'
        ), $camNr);

        if (!isset($cams_params[$camNr])) {
            throw new \Exception('cameraNumber not found');
        }
        $camData = $cams_params[$camNr];
        if (@empty($camData['onvif_profile_token']['v'])) {
            throw new \Exception('onvif_profile_token not set');
        }

        if (!@empty($camData['onvif_http_port']['v'])) {
            $onvif_http_port = $camData['onvif_http_port']['v'];
        } elseif (!@empty($camData['InetCam_http_port']['v'])) {
            $onvif_http_port = $camData['InetCam_http_port']['v'];
        } else {
            $onvif_http_port = 80;
        }

        $connectionData = array(
            'host' => $camData['InetCam_IP']['v'],
            'port' => $onvif_http_port,
            'profile_token' => $camData['onvif_profile_token']['v']
        );
        if (!@empty($camData['InetCam_USER']['v'])) {
            $connectionData['username'] = $camData['InetCam_USER']['v'];
            if (!@empty($camData['InetCam_PASSWD']['v'])) {
                $connectionData['password'] = $camData['InetCam_PASSWD']['v'];
            } else {
                $connectionData['password'] = '';
            }
        }
        if (@!empty($requestData['sid'])) {
            $connectionData['sid'] = (string)$requestData['sid'];
        }

        $connectionData['cameraNumber'] = $camNr;

        $this->connectionData = $connectionData;
        return $this->connectionData;
    } /* getConnectionData() */

    /**
     * Extract speed component from POST request to be used in SOAP.
     * @param array $data
     * @param boolean $expandPanTilt - set x/y = 0 if pan/tilt speed is not set
     * @return array
     */
    protected function getSpeedVector($data = array(), $expandPanTilt = false)
    {
        // collect speed data
        $speed = array();

        if (isset($data['speedPan']) || isset($data['speedTilt'])) {
            $speed['PanTilt'] = array();

            if (isset($data['speedPan'])) {
                $speed['PanTilt']['x'] = $data['speedPan'];
            } elseif ($expandPanTilt) {
                $speed['PanTilt']['x'] = 0;
            }

            if (isset($data['speedTilt'])) {
                $speed['PanTilt']['y'] = $data['speedTilt'];
            } elseif ($expandPanTilt) {
                $speed['PanTilt']['y'] = 0;
            }
        }

        if (isset($data['speedZoom'])) {
            $speed['Zoom'] = array('x' => $data['speedZoom']);
        }

        return $speed;
    }

    public function getPtzSpaces($data = array())
    {
        // TODO do real onvif query
        $this->success(array(
            'coordSpaces' => array(
                'zoom' => array(
                    'min' => 0,
                    'max' => 1
                ),
                'pan' => array(
                    'min' => -1,
                    'max' => 1
                ),
                'tilt' => array(
                    'min' => -1,
                    'max' => 1
                )
            ),
            'speedSpaces' => array(
                'zoom' => array(
                    'min' => 0,
                    'max' => 1
                ),
                'pan' => array(
                    'min' => 0,
                    'max' => 1
                ),
                'tilt' => array(
                    'min' => 0,
                    'max' => 1
                )
            )
        ));
    }

    public function getPtzStatus($data = array())
    {
        $this->getConnectionData($data);

        if ($this->doPtzCommand('GetStatus', array('ProfileToken' => $this->connectionData['profile_token']))) {
            $pos = $this->res['result']->PTZStatus->Position;
            $this->success(
                array(
                    'position' => array(
                        // convert from possible scientific notation to dot notation
                        'pan' => sprintf('%F', $pos->PanTilt->x),
                        'tilt' => sprintf('%F', $pos->PanTilt->y),
                        'zoom' => sprintf('%F', $pos->Zoom->x)
                    )
                )
            );
        }
    } /* getPtzStatus() */

    public function getPtzPresets($data = array())
    {
        $this->getConnectionData($data);

        if ($this->doPtzCommand('GetPresets', array('ProfileToken' => $this->connectionData['profile_token']))) {
            $result = array();
            if (property_exists($this->res['result'], 'Preset')) {
                $presets = $this->res['result']->Preset;
                if (!is_array($presets)) {
                    $presets = array($presets);
                }
                foreach ($presets as $preset) {
                    $presetInfo = array();
                    $presetInfo['name'] = $preset->Name;
                    $presetInfo['token'] = $preset->token;
                    if (property_exists($preset, 'PTZPosition')) {
                        $x = $preset->PTZPosition->PanTilt->x;
                        $y = $preset->PTZPosition->PanTilt->y;
                        $z = $preset->PTZPosition->Zoom->x;
                    } else {
                        $x = $y = $z = 0;
                    }
                    $presetInfo['position'] = array(
                        // convert from possible scientific notation to dot notation
                        'pan' => sprintf('%F', $x),
                        'tilt' => sprintf('%F', $y),
                        'zoom' => sprintf('%F', $z)
                    );
                    $result[] = $presetInfo;
                }
            }
            $this->success($result);
        }
    } /* getPtzPresets() */

    public function moveAbsolute($data = array())
    {
        if (!isset($data['pan']) && !isset($data['tilt']) && !isset($data['zoom'])) {
            throw new \Exception('Position not set');
        }

        $this->getConnectionData($data);

        // collect move parameters
        $position = array(
            'PanTilt' => array(),
            'Zoom' => array()
        );

        isset($data['pan']) ? $position['PanTilt']['x'] = $data['pan'] : '';
        isset($data['tilt']) ? $position['PanTilt']['y'] = $data['tilt'] : '';
        isset($data['zoom']) ? $position['Zoom']['x'] = $data['zoom'] : '';

        $speed = $this->getSpeedVector($data);

        // do the request
        $requestParams = array(
            'Position' => $position,
            'Speed' => $speed,
            'ProfileToken' => $this->connectionData['profile_token']
        );

        if (!empty($speed)) {
            $requestParams['Speed'] = $speed;
        }

        if ($this->doPtzCommand('AbsoluteMove', $requestParams)) {
            $this->success();
        }
    } /* moveAbsolute() */

    public function moveContinuous($data = array())
    {
        if (!isset($data['speedPan']) && !isset($data['speedTilt']) && !isset($data['speedZoom'])) {
            throw new \Exception('Move direction not set');
        }

        $this->getConnectionData($data);

        // collect move parameters
        $velocity = $this->getSpeedVector($data, true);

        // do the request
        $requestParams = array(
            'Velocity' => $velocity,
            'Timeout' => 'PT30S',
            'ProfileToken' => $this->connectionData['profile_token']
        );

        if (!empty($speed)) {
            $requestParams['Speed'] = $speed;
        }

        if ($this->doPtzCommand('ContinuousMove', $requestParams)) {
            $this->success();
        }
    } /* moveContinuous() */

    public function moveStop($data = array())
    {
        $this->getConnectionData($data);

        if ($this->doPtzCommand(
            'Stop',
            array(
                'ProfileToken' => $this->connectionData['profile_token'],
                'PanTilt' => true,
                'Zoom' => true
            )
        )) {
            $this->success();
        }
    } /* moveStop() */

    public function gotoPreset($data = array())
    {
        if (!isset($data['presetToken'])) {
            throw new \Exception('presetToken not set');
        }

        $this->getConnectionData($data);

        $speed = $this->getSpeedVector($data);

        $requestParams = array(
            'PresetToken' => $data['presetToken'],
            'ProfileToken' => $this->connectionData['profile_token']
        );

        if (!empty($speed)) {
            $requestParams['Speed'] = $speed;
        }

        if ($this->doPtzCommand('GotoPreset', $requestParams)) {
            $this->success();
        }
    } /* gotoPreset() */

    public function gotoHomePosition($data = array())
    {
        $this->getConnectionData($data);

        $speed = $this->getSpeedVector($data);

        $requestParams = array(
            'Speed' => $speed,
            'ProfileToken' => $this->connectionData['profile_token']
        );

        if (!empty($speed)) {
            $requestParams['Speed'] = $speed;
        }

        if ($this->doPtzCommand('GotoHomePosition', $requestParams)) {
            $this->success();
        }
    } /* gotoHomePosition() */

    public function setHomePosition($data = array())
    {
        $this->getConnectionData($data);

        if ($this->doPtzCommand(
            'SetHomePosition',
            array('ProfileToken' => $this->connectionData['profile_token'])
        )) {
            $this->success();
        }
    } /* setHomePosition() */


    public function setPreset($data = array())
    {
        if (!isset($data['presetName'])) {
            throw new \Exception('presetName not set');
        }

        $this->getConnectionData($data);

        if ($this->doPtzCommand(
            'SetPreset',
            array(
                'ProfileToken' => $this->connectionData['profile_token'],
                'PresetName' => $data['presetName']
            )
        )) {
            $this->success();
        }
    } /* setPreset() */

    public function removePreset($data = array())
    {
        if (!isset($data['presetToken'])) {
            throw new \Exception('presetToken not set');
        }

        $this->getConnectionData($data);

        if ($this->doPtzCommand(
            'RemovePreset',
            array('ProfileToken' => $this->connectionData['profile_token'],
            'PresetToken' => $data['presetToken'])
        )) {
            $this->success();
        }
    } /* removePreset() */

    protected function doPtzCommand($cmd, $data)
    {
        $this->res = $this->doRequest($this->connectionData, 'ptz', $cmd, $data);

        if ($this->res['isOk']) {
            return true;
        } else {
            $this->error($cmd . ' failed: ' . $this->res['error_message']);
            return false;
        }
    } /* doPtzCommand() */
}

$controller = new OnvifPtzController();
