<?php
/**
 * Copyright (C) 2004-2021 "Setevye informacionnye sistemy", Ltd (Russia)
 *                          en: "Network Information Systems"
 */

/**
 *
 * @file lib/adb.php
 * @brief В файле реализован класс, который обеспечивает взаимодействие с БД,<br />а также инициализируется экземпляр
 * этого класса
 *
 * Все обращения к БД должны быть реализованы посредством этого класса,
 * а работа с БД в обход этого класса крайне нежелательна.
 *
 * Для инициализации экземпляра класса используется объект конфигурации $conf из /etc/avreg/site-defaults.php
 *
 * Для подключения к БД класс использует PDO
 *
 */

namespace Avreg;

use \PDO;
use \PDOException;

if (empty($non_config)) {
    require_once 'config.inc.php';
}

/// Инициализируем класс по работе с БД
$adb = new Adb($conf);

/**
 * @class Adb
 * @brief Класс взаимодействия с БД
 *
 */
class Adb
{
    const SQLDATETIMEFMT = 'Y-m-d H:i:s';

    // Название БД
    private $database = '';
    /// Пользователь БД
    private $user = '';
    /// Пароль БД
    private $password = '';
    /// Тип БД mysql - MySql, pgsql - PostgreSql
    private $dbAdapter = 'mysql';
    /// Хост БД
    private $host = null;
    /// Объект для работы с БД
    private $dbh = false;

    /**
     *  Конструктор по умолчанию
     * Устанавливает соединение с БД
     * @params array $params масив конфигурации класса, передаётся авреговый $conf
     * @return \Avreg\Adb если соединение с базой успешно, false если произошла ошибка.
     */
    public function __construct($params)
    {
        $this->database = $params['db-name'];
        $this->user = $params['db-user'];
        $this->password = $params['db-passwd'];
        if (isset($params['db-type']) && !empty($params['db-type'])) {
            $this->dbAdapter = $params['db-type'];
        }
        if (isset($params['db-host']) && !empty($params['db-host'])) {
            $this->host = $params['db-host'];
        }

        if (@!empty($params['db-port'])) {
            $dsn = "{$this->dbAdapter}:host={$this->host};port={$params['db-port']};dbname={$this->database}";
        } else {
            $dsn = "{$this->dbAdapter}:host={$this->host};dbname={$this->database}";
        }

        try {
            $this->dbh = new PDO(
                $dsn,
                $this->user, $this->password,
                array(
                    PDO::ATTR_PERSISTENT => true,
                    PDO::ATTR_ERRMODE => PDO::ERRMODE_WARNING // PDO::ERRMODE_SILENT
                )
            );
        } catch (PDOException $e) {
            trigger_error($e->getMessage());
            throw $e;
        }

        if ($this->dbAdapter === 'mysql') {
            $res = $this->dbh->exec("SET NAMES 'utf8' COLLATE 'utf8_general_ci'");
        } else {
            $res = $this->dbh->exec("SET NAMES 'utf8'");
        }
        if ($res === false) {
            $this->checkError();
        }
        return true;
    }

    /**
     *  Деструктор по умолчанию
     * Закрывает соединение с БД
     */
    public function __destruct()
    {
        $this->dbh = null;
    }

    public function lastError($stmt, &$msgRef)
    {
        $errInfo = $stmt ? $stmt->errorInfo() : $this->dbh->errorInfo();
        if ($errInfo[0] === '00000') {
            return false;
        } else {
            $msgRef = implode('; ', $errInfo);
            return true;
        }
    }

    /**
     *  Проверка на ошибку в запросе к БД.
     *
     * @param bool $die true - закончить скрипт, false - вывести ошибку
     * @param object $stmt PDOStatement if used by last query.
     *
     * @return true - если ошибка, false - если нет ошибок
     */
    public function checkError($die = true, $stfm = null)
    {
        if ($this->lastError($stfm, $msgRef)) {
            // $ui = preg_replace('!://(.*)@!', '://{USER}:********@', $r->getUserInfo());
            if ($die) {
                @header('Content-Type: text/plain; charset=' . $GLOBALS['chset'], true, 500);
                echo 'DBMS/User Message: ' . $msgRef;
            }

            // echo $r->getDebugInfo();

            if ($die) {
                trigger_error($msgRef);
                throw new \Exception($msgRef);
            }

            return true;
        } else {
            return false;
        }
    }

    /**
     *  Метод позволяет получить события по указанным параметрам
     *
     *
     * @param array $param Параметры
     * - $param['events'] тип событий событий (изображения, аудио, видео)
     * - $param['cameras']  список камер
     * - $param['date'] дата событий c часовой градацией
     * - $param['to']   точные ограничители DT1
     * - $param['from']
     * - $param['limit']
     * - $param['offset']
     *
     * @return array масив событий
     */

    public function galleryGetEvent($param)
    {
        $events = array();
        $query = "select " . $this->dateFormat('DT1') . ", DT1, EVT_CONT, ALT2, ALT1, CAM_NR, FILESZ_KB, EVT_ID, " .
            $this->timediff('DT1', 'DT2') . ", DT2";
        $query .= ' from EVENTS';
        $query .= ' where ' . $this->whereIntColumnValue("CAM_NR", $param['cameras']);
        $query .= ' and ' . $this->whereIntColumnValue("EVT_ID", $param['events']);

        if (!@empty($param['from'])) {
            $query .= " and DT1 >= '" . $param['from'] . "'";
        }
        if (!@empty($param['to'])) {
            $query .= " and DT1 <= '" . $param['to'] . "'";
        }

        if (isset($param['date']) && is_array($param['date'])) {
            $query .= ' ' . $this->dateRange($param['date'], 'DT1');
        }

        // сортировать по дате, от текущей позиции с лимитом заданный в конфиге
        $query .= ' order by DT1,CAM_NR limit ' . $param['limit'] . ' offset ' . $param['offset'];

        foreach ($this->dbh->query($query, PDO::FETCH_NUM) as $row) {
            $row[6] = filesizeHuman($row[6]);
            if (in_array((int)$row[7], array(15, 16, 17, 18, 19, 20, 21))) {
                $row[7] = 'image';
            } elseif ((int)$row[7] === 23) {
                $row[7] = 'video';
            } elseif ((int)$row[7] === 12) {
                $row[7] = 'video';
            } elseif ((int)$row[7] === 32) {
                $row[7] = 'audio';
            }

            // формирование уникального индекса, для работы кэша в браузере пользователя
            // TODO слишком длинный ключ
            $events[str_replace(array('/', '.'), '_', $row[5] . '_' . $row[2] . '_' . $row[0])] = $row;
        }
        $this->checkError();

        return $events;
    }

    /**
     *  Метод позволяет получить дату текущего события
     *  FIXME вроде исп. для навигации по скролу (всплывающее окно)
     *
     * @param array $param Параметры
     * - $param['events'] тип событий событий (изображения, аудио, видео)
     * - $param['cameras']  список камер
     * - $param['date'] дата событий
     * - $param['limit']
     * - $param['offset']
     *
     * @return array масив событий
     */

    public function galleryGetEventDate($param)
    {
        $events = array();
        $query = "select " . $this->dateFormat('DT2') . ", DT2";
        $query .= ' from EVENTS';
        $query .= ' where ' . $this->whereIntColumnValue("CAM_NR", $param['cameras']);
        $query .= ' and ' . $this->whereIntColumnValue("EVT_ID", $param['events']);

        if (!@empty($param['from'])) {
            $query .= " and DT1 >= '" . $param['from'] . "'";
        }
        if (!@empty($param['to'])) {
            $query .= " and DT1 <= '" . $param['to'] . "'";
        }

        if (isset($param['date']) && is_array($param['date'])) {
            $query .= ' ' . $this->dateRange($param['date'], 'DT1');
        }

        // сортировать по дате, от текущей позиции с лимитом заданный в конфиге
        $query .= ' order by DT1,CAM_NR limit ' . $param['limit'] . ' offset ' . $param['offset'];

        $stmt = $this->dbh->query($query);
        if ($stmt === false) {
            $this->checkError();
        }

        $row = $stmt->fetch();
        if ($row === false) {
            $this->checkError(true, $stmt);
        }

        $date_events = $row[1];
        $stmt->closeCursor();
        unset($stmt);

        return $date_events;
    }

    /**
     *  Метод позволяет получить статистику о файловых событиях в таблицe EVENTS
     *
     * @param array $param Параметры
     * - $params['cameras']  список камер
     * - $params['from']     дата from (from >= DT1)
     * - $params['to']       дата to (to <= DT1)
     * @return array         with fields: files, latest and oldest
     */
    public function galleryEventsGetStat($params = array())
    {
        $stat = array(
            'files' => 0,
            'latest' => '1970-01-01 00:00:00',
            'oldest' => '1970-01-01 00:00:00'
        );

        $query = 'select COUNT(*) as files, MAX(DT1) as oldest, MIN(DT1) as latest from EVENTS';
        $query .= ' where EVT_ID in (12, 15,16,17, 23, 32)';
        if (!empty($params)) {
            foreach ($params as $key => $value) {
                if (empty($value)) {
                    continue;
                }
                switch ($key) {
                    case 'cameras':
                        $query .= ' and ' . $this->whereIntColumnValue("CAM_NR", $value);
                        break;
                    case 'from':
                        $query .= " and DT1 >= '$value'";
                        break;
                    case 'to':
                        $query .= " and DT1 <= '$value'";
                        break;
                    case 'method':
                    case 'initially':
                        break;
                    default:
                        die("galleryEventsGetStat() failed: unknown param \"$key\" = \"$value\"");
                }
            }
            unset($key, $value);
        }

        $stmt = $this->dbh->query($query);
        if ($stmt === false) {
            $this->checkError();
        }

        $row = $stmt->fetch();
        if ($row === false) {
            $this->checkError(true, $stmt);
        }

        if (!empty($row[0])) {
            $stat['files'] = (int)$row[0];
            $stat['latest'] = $row[1];
            $stat['oldest'] = $row[2];
        }

        $stmt->closeCursor();
        unset($stmt);

        return $stat;
    } /* galleryEventsGetStat() */

    /**
     *  Метод позволяет получить статистику о дереве событий
     *
     * @param array $param Параметры
     * - $params['cameras']  список камер
     * - $params['from']     дата from (from >= DT1_MIN)
     * - $params['to']       дата to (to <= DT1_MAX)
     * @return array         with fields: files, latest_update and oldest_update
     */
    public function galleryTreeEventsGetStat($params = array())
    {
        $stat = array(
            'files' => 0,
            'latest_update' => '1970-01-01 00:00:00',
            'oldest_update' => '1970-01-01 00:00:00'
        );

        $query = 'select SUM(IMAGE_COUNT+VIDEO_COUNT+AUDIO_COUNT) as files,';
        $query .= ' MAX(DT1_MAX) as latest_update, MIN(DT1_MIN) as oldest_update';
        $query .= ' from TREE_EVENTS';
        if (!empty($params)) {
            $query .= ' where 1=1';

            foreach ($params as $key => $value) {
                if (empty($value)) {
                    continue;
                }
                switch ($key) {
                    case 'cameras':
                        $query .= ' and ' . $this->whereIntColumnValue("CAM_NR", $value);
                        break;
                    case 'from':
                        $query .= " and DT1_MIN >= '$value'";
                        break;
                    case 'to':
                        $query .= " and DT1_MAX <= '$value'";
                        break;
                    case 'method':
                    case 'initially':
                        break;
                    default:
                        die("galleryTreeEventGetStat() failed: unknown param \"$key\" = \"$value\"");
                }
            }
            unset($key, $value);
        }

        $stmt = $this->dbh->query($query);
        if ($stmt === false) {
            $this->checkError();
        }

        $row = $stmt->fetch();
        if ($row === false) {
            $this->checkError(true, $stmt);
        }

        if (!empty($row[0])) {
            $stat['files'] = (int)$row[0];
            $stat['latest_update'] = $row[1];
            $stat['oldest_update'] = $row[2];
        }

        $stmt->closeCursor();
        unset($stmt);

        return $stat;
    } /* galleryTreeEventGetStat() */


    /***
     *  Метод позволяет получить дерево событий
     *
     * @param array $param Параметры
     * - $params['cameras']  список камер
     * - $params['from']     дата from (from >= DT1_MIN)
     * - $params['to']       дата to (to <= DT1_MAX)
     * @return array хеш-масив дерева событий со почасовой статистикой с разбиением на камеры
     *      {
     *          "YYYY_MM_DD_HH" = {
     *              "cNN" => { // NN - номер камеры
     *                  "is" = > image_size,
     *                  "ic" => image_count,
     *                  "vs" = > video_size,
     *                  "vc" => video_count,
     *                  "as" = > audio_size,
     *                  "ac" => audio_count
    *
     *              },
     *              ...
     *          },
     *          ...
     *      }
     */
    public function galleryGetTreeEvents($params)
    {
        $tree_events_result = array();

        $query = 'select * from TREE_EVENTS';
        if (!empty($params)) {
            $query .= ' where 1=1';

            foreach ($params as $key => $value) {
                if (empty($value)) {
                    continue;
                }
                switch ($key) {
                    case 'cameras':
                        $query .= ' and ' . $this->whereIntColumnValue("CAM_NR", $value);
                        break;
                    case 'from':
                        $query .= " and DT1_MIN >= '$value'";
                        break;
                    case 'to':
                        $query .= " and DT1_MAX <= '$value'";
                        break;
                    case 'method':
                    case 'initially':
                        break;
                    default:
                        die("galleryGetTreeEvents() failed: unknown param \"$key\" = \"$value\"");
                }
            }
            unset($key, $value);
        }
        $query .= ' order by DT1_MAX desc';

        $stmt = $this->dbh->prepare($query, array(PDO::ATTR_CURSOR => PDO::CURSOR_FWDONLY));
        if ($stmt === false) {
            $this->checkError(true, $stmt);
        }

        if (false === $stmt->execute()) {
            $this->checkError(true, $stmt);
        }
        /*
        $stmt = $this->dbh->query($query);
        if ($stmt === false) {
            $this->checkError();
        }
        */

        while ($row = $stmt->fetch(PDO::FETCH_NAMED)) {
            $hourKey = $row[$this->key('BYHOUR')];
            $camKey = 'c' . $row[$this->key('CAM_NR')];
            if (!isset($tree_events_result[$hourKey])) {
                $tree_events_result[$hourKey] = array();
            }

            $camStatRef = &$tree_events_result[$hourKey][$camKey];
            $tree_events_result[$hourKey][$camKey] = array();

            $video_count = intval($row[$this->key('VIDEO_COUNT')]);
            $audio_count = intval($row[$this->key('AUDIO_COUNT')]);
            $image_count = intval($row[$this->key('IMAGE_COUNT')]);

            if ($video_count) {
                $camStatRef['vc'] = $video_count;
                $camStatRef['vs'] = intval($row[$this->key('VIDEO_SIZE')]);
            }
            if ($audio_count) {
                $camStatRef['ac'] = $audio_count;
                $camStatRef['as'] = intval($row[$this->key('AUDIO_SIZE')]);
            }
            if ($image_count) {
                $camStatRef['ic'] = $image_count;
                $camStatRef['is'] = intval($row[$this->key('IMAGE_SIZE')]);
            }
        }
        if ($row === false) {
            $this->checkError(true, $stmt);
        }
        unset($stmt);

        return $tree_events_result;
    } /* galleryGetTreeEvents() */

    /**
     *  Метод обновляет дерево событий
     * @param string $start_hb Дата начала обновления дерева, выровненное по часу
     * @param string $end_hb Дата окончания обновления дерева, выровненное по часу
     * @param string $to Ограничитель DT1 для предотвращения зацикливания
     *                    при постоянном добавлении в EVENTS и при этом неопр. $start_hb и $end_hb
     * @param bool|string $cameras Список камер
     * @return array
     */
    public function galleryUpdateTreeEvents($start_hb, $end_hb, $to = false, $cameras = false)
    {
       $query = 'SELECT DT1, CAM_NR, EVT_ID, FILESZ_KB FROM EVENTS WHERE EVT_ID IN (12, 15,16,17, 23, 32)';
        if ($start_hb) {
            $hour_start = date('Y-m-d H:00:00', strtotime($start_hb));
            $query .= " AND DT1 >= '$hour_start'";
        }
        if ($end_hb) {
            $hour_end = date('Y-m-d H:59:59', strtotime($end_hb));
            $query .= " AND DT1 <= '$hour_end'";
        }
        if ($to) {
            $query .= " AND DT1 <= '$to'";
        }
        if ($cameras) {
            $query .= " AND " . $this->whereIntColumnValue("CAM_NR", $cameras);
        }
        $query .= ' ORDER BY DT1';

        /*
         *  DATE CAM_NR DT1_MAX DT1_MIN IMAGE_COUNT IMAGE_SIZE VIDEO_COUNT VIDEO_SIZE AUDIO_COUNT AUDIO_SIZE
         *   0      1      2       3         4           5          6           7          8           9
         */
        $treeEvents = array();

        foreach($this->dbh->query($query) as $row) {
            /* XXX date() + strtotime() - very slow */
            $date = strtr(substr($row[$this->key('DT1')], 0, 13), '- ', '__');
            $key = $date . '_' . $row[$this->key('CAM_NR')];

            $row_ref = &$treeEvents[$key];
            if (!$row_ref) {
                $treeEvents[$key] = array(
                    $date,
                    $row[$this->key('CAM_NR')],
                    $row[$this->key('DT1')],
                    $row[$this->key('DT1')],
                    0,0,0,0,0,0
                );
                $row_ref = &$treeEvents[$key];
            }
            $evt_id = (int)$row[$this->key('EVT_ID')];
            switch ($evt_id) {
                case 15:
                case 16:
                case 17:
                    $row_ref[4] += 1;
                    $row_ref[5] += (int)$row[$this->key('FILESZ_KB')];
                    break;
                case 12:
                case 23:
                    $row_ref[6] += 1;
                    $row_ref[7] += (int)$row[$this->key('FILESZ_KB')];
                    break;
                case 32:
                    $row_ref[8] +=1;
                    $row_ref[9] += (int)$row[$this->key('FILESZ_KB')];
                    break;
            }
            if ($row[$this->key('DT1')] > $row_ref[2]) {
                $row_ref[2] = $row[$this->key('DT1')];
            }
            if ($row[$this->key('DT1')] < $row_ref[3]) {
                $row_ref[3] = $row[$this->key('DT1')];
            }
        }
        $this->checkError();

        $query = 'DELETE FROM TREE_EVENTS WHERE 1=1';
        if ($start_hb) {
            $query .= " AND BYHOUR >= '" . date('Y_m_d_H', strtotime($start_hb)) . "'";
        }
        if ($end_hb) {
            $query .= " AND BYHOUR <= '" . date('Y_m_d_H', strtotime($end_hb)) . "'";
        }
        if ($to) {
            $query .= " AND DT1_MAX <= '$to'";
        }
        if ($cameras) {
            $query .= ' AND ' . $this->whereIntColumnValue("CAM_NR", $cameras);
        }

        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }

        if (false === $this->dbh->exec('BEGIN')) {
            $this->checkError();
        }

        $query = 'INSERT INTO TREE_EVENTS';
        $query .= ' (BYHOUR, CAM_NR, DT1_MAX, DT1_MIN, ';
        $query .= 'IMAGE_COUNT, IMAGE_SIZE, VIDEO_COUNT, VIDEO_SIZE, AUDIO_COUNT, AUDIO_SIZE)';
        $query .= ' VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?)';

        $stmt = $this->dbh->prepare($query, array(PDO::ATTR_CURSOR => PDO::CURSOR_FWDONLY));
        if ($stmt === false) {
            $this->checkError();
        }

        foreach ($treeEvents as &$row) {
            if (!$stmt->execute($row)) {
                // FIXME ROLLBACK ???
                $this->checkError(true, $stmt);
            }
        }
        $this->checkError();

        if (false === $this->dbh->exec('COMMIT')) {
            $this->checkError();
        }

        return array('status' => 'success');
    }

    /**
     *  Метод получения событий
     * @param int $camera номер камеры
     * @param int $ser_nr
     * @param string $timebegin дата начала
     * @param bool|string $timeend дата окончания
     * @param string $order сортировка
     *
     * @return array масив событий
     */

    public function getSnapshots($camera, $ser_nr, $timebegin, $timeend = false, $order = '')
    {
        $files = array();
        $query = 'SELECT ' . $this->datePart('timestamp', 'DT1') . ' as START, '
            . $this->datePart('timestamp', 'DT1') . ' as FINISH, '
            . '  EVT_ID, FILESZ_KB, FRAMES, ALT1 as U16_1, ALT2 as U16_2, EVT_CONT' . "\n"
            . 'FROM EVENTS' . "\n"
            . "WHERE CAM_NR=$camera\n"
            . "  AND EVT_ID >= 15 AND EVT_ID <= 21\n";
        if (empty($timeend)) {
            $query .= "  AND SESS_NR = $ser_nr\n";
            $query .= "  AND DT1 >= '$timebegin'\n";
        } else {
            $query .= "  AND (DT1 between '$timebegin' and '$timeend')\n";
        }
        $query .= "ORDER BY DT1,CAM_NR " . $order;

        foreach ($this->dbh->query($query) as $row) {
            $f = array();
            foreach ($row as $k => $v) {
                $k = strtoupper($k);
                $f[$k] = trim($v);
            }
            $files[] = $f;
        }

        $this->checkError();

        return $files;
    }

    /**
     *  Метод позволяет получить статистику сохранённых снапшотов за сеанс записи.
     * @param string $cams_csv список камер
     * @param string $timebegin дата начала
     * @param string $timeend дата окончания
     * @param string $order сортировка
     *
     * @return array масив событий
     */
    public function getSnapStatsByRecSeries($cams_csv, $timebegin, $timeend, $order = '')
    {

        $files = array();
        $query = 'SELECT ' . $this->datePart('timestamp', 'E1.DT1') . ' as START, '
            . $this->datePart('timestamp', 'E2.DT1') . ' as FINISH, '
            . '  E1.CAM_NR, E1.SESS_NR AS SESS_NR,' . "\n"
            . '  count(E3.DT1) as SHANPHOTS_NB' . "\n"
            . "FROM EVENTS AS E1\n"
            . '  JOIN EVENTS AS E2 ON (E1.SESS_NR = E2.SESS_NR AND E1.CAM_NR = E2.CAM_NR AND'
            . '    E1.DT1 = E2.DT2 AND E1.EVT_ID = 13 AND E2.EVT_ID = 14)' . "\n"
            . '  JOIN EVENTS AS E3 ON (E1.SESS_NR = E3.SESS_NR AND E1.CAM_NR = E3.CAM_NR)' . "\n"
            . "WHERE E1.CAM_NR in ($cams_csv)\n"
            . "  AND E1.EVT_ID in (13)\n"
            . "  AND ((E1.DT1 between '$timebegin' and '$timeend') and "
            . "      (E2.DT1 is null or E2.DT1 between '$timebegin' and '$timeend'))\n"
            . '  AND (E3.EVT_ID >= 15 and E3.EVT_ID <= 21 and E3.DT1 >= E1.DT1 and E3.DT1 <= E2.DT1)' . "\n"
            . 'GROUP BY E1.DT1, E2.DT1, E1.CAM_NR, E1.SESS_NR' . "\n"
            . "ORDER BY E2.DT1 " . $order;

        foreach ($this->dbh->query($query) as $row) {
            $f = array();
            foreach ($row as $k => $v) {
                $k = strtoupper($k);
                $f[$k] = trim($v);
            }
            $files[] = $f;
        }
        $this->checkError();

        return $files;
    }

    /**
     *  Метод позволяет получить статистику по сохранённым снапшотам за интервал.
     * @param string $cams_csv список камер
     * @param string $timebegin дата начала
     * @param string $timeend дата окончания
     * @param string $order сортировка
     *
     * @return array масив событий
     */
    public function getSnapStatsByInterval($cams_csv, $timebegin, $timeend, $order = '')
    {
        $files = array();
        $query = 'select ' . "\n"
           . $this->datePart('timestamp', 'min(DT1)') . " as START,\n"
           . $this->datePart('timestamp', 'max(DT1)') . " as FINISH,\n"
           . "CAM_NR, '0' as SESS_NR, count(*) as SHANPHOTS_NB\n"
           . "from EVENTS\n"
           . "where\n"
           . "(DT1 between '$timebegin' and '$timeend')\n"
           . "and CAM_NR in ($cams_csv)\n"
           . "and EVT_ID >= 15 and EVT_ID <= 21\n"
           . "group by CAM_NR\n"
           . 'order by max(DT1) ' . $order;

        foreach ($this->dbh->query($query) as $row) {
            $f = array();
            foreach ($row as $k => $v) {
                $k = strtoupper($k);
                $f[$k] = trim($v);
            }
            $files[] = $f;
        }
        $this->checkError();
        return $files;
    }

    /**
     *  Метод позволяет получить масив событий для offline модуля
     * @param array $cams список камер
     * @param array $dateTimeRange дата
     * @param array $event_ids массив с типами событий
     * @param array $weekdays дни недели
     * @param bool $time_slices_mode
     * @param array|bool $page лимит и номер страницы
     *
     * @return array масив событий
     */
    public function eventsSelect($cams, $dateTimeRange, $event_ids, $weekdays, $time_slices_mode = false, $page = false)
    {
        global $__EVENTS_QUERY_INFO; // for _playlist.php

        $req_continuos_events = false;
        $req_only_media_events = false;
        if ($event_ids && is_array($event_ids)) {
            array_map('intval', $event_ids);
            sort($event_ids, SORT_NUMERIC);

            // XXX must sync to $env_id_ar from lang/*/*/common.inc.php
            if (12 === count(array_intersect(array(1,3,4,5,12,13,14,15,16,17,23,32), $event_ids))) {
                // all events
                $event_ids = false;
                $req_continuos_events = true;
            } else {
                if (!empty(array_intersect(array(12, 13, 14, 23, 32), $event_ids))) {
                    $req_continuos_events = true;
                }
                if (3 === count(array_intersect(array(12, 23, 32), $event_ids))) {
                    $req_only_media_events = true;
                }
            }
        } else {
            $event_ids = false;
        }

        if ($weekdays && is_array($weekdays)) {
            array_map('intval', $weekdays);
            sort($weekdays, SORT_NUMERIC);
            if (7 === count(array_intersect(array(0,1,2,3,4,5,6), $weekdays))) {
                $weekdays = false;
            }
        } else {
            $weekdays = false;
        }

        $events = array();
        $query = 'SELECT ' . $this->datePart('timestamp', 'DT1') . ' as UDT1, ' . $this->datePart('timestamp', 'DT2') .
            ' as UDT2,';
        $query .= ' CAM_NR, EVT_ID, SESS_NR AS SER_NR, FILESZ_KB, FRAMES, ALT1, ALT2, EVT_CONT';
        $query .= ' FROM EVENTS';
        $query .= ' WHERE ';

        if (!$time_slices_mode) {
            // сплошной отрезок времени
            $timebegin = sprintf(
                '%04u-%02u-%02u %02u:%02u:00',
                $dateTimeRange['from'][0],
                $dateTimeRange['from'][1],
                $dateTimeRange['from'][2],
                $dateTimeRange['from'][3],
                $dateTimeRange['from'][4]
            );
            $timeend = sprintf(
                '%04u-%02u-%02u %02u:%02u:59',
                $dateTimeRange['to'][0],
                $dateTimeRange['to'][1],
                $dateTimeRange['to'][2],
                $dateTimeRange['to'][3],
                $dateTimeRange['to'][4]
            );

            if ($req_continuos_events) {
                $query .= '(';
            }
            $query .= "(DT1 BETWEEN '$timebegin' AND '$timeend')";

            if ($req_continuos_events) {
                $query .= " OR (DT2 BETWEEN '$timebegin' AND '$timeend'))";
            }
        } else {
            // интервалы
            $timebegin = sprintf(
                '%04u-%02u-%02u 00:00:00',
                $dateTimeRange['from'][0],
                $dateTimeRange['from'][1],
                $dateTimeRange['from'][2]
            );
            $timeend = sprintf(
                '%04u-%02u-%02u 23:59:59',
                $dateTimeRange['to'][0],
                $dateTimeRange['to'][1],
                $dateTimeRange['to'][2]
            );
            $time_in_day_begin = sprintf('%02u:%02u:00', $dateTimeRange['from'][3], $dateTimeRange['from'][4]);
            $time_in_day_end   = sprintf('%02u:%02u:59', $dateTimeRange['to'][3], $dateTimeRange['to'][4]);

            if ($req_continuos_events) {
                $query .= '((';
            }
            $query .= "(DT1 BETWEEN '$timebegin' AND '$timeend')";
                if ($weekdays) {
                $query .= ' AND (' . $this->datePart('weekday', 'DT1') . ' IN (' . implode(',', $weekdays) . '))';
            }
            $query .= ' AND (' . $this->datePart('time', 'DT1') . " BETWEEN '$time_in_day_begin' AND '$time_in_day_end')";
            if ($req_continuos_events) {
                $query .= ') OR (';
                $query .= "(DT2 BETWEEN '$timebegin' AND '$timeend')";
                if ($weekdays) {
                    $query .= ' AND (' . $this->datePart('weekday', 'DT2') . ' IN (' . implode(',', $weekdays) . '))';
                }
                $query .= ' AND (' . $this->datePart('time', 'DT2') . " BETWEEN '$time_in_day_begin' AND '$time_in_day_end')";
            }
            if ($req_continuos_events) {
                $query .= '))';
            }
        }

        if (!@empty($cams)) {
            $query .= ' AND CAM_NR IN (0, ' . implode(',', $cams) . ')';
        }

        if (!@empty($event_ids)) {
            $query .= ' AND EVT_ID IN (' . implode(',', $event_ids) . ')';
        }

        if ($req_only_media_events) {
            $query .= ' ORDER BY DT2'; // file started
        } else {
            $query .= ' ORDER BY DT1'; // event date = file closing (avregd 6.3)
        }

        if ($req_continuos_events) {
            if ($req_only_media_events) {
                $query .= ',DT1'; // need for MySQL perfomance
            } else {
                $query .= ',DT2'; // need for MySQL perfomance
            }
        }

        if (!empty($page)) {
            $query .= ',CAM_NR'; // order by DT1,CAM_NR if limit
            $query .= ' LIMIT ' . $page['limit'];
            $query .= ' OFFSET ' . $page['offset'];
        }

        // error_log($query);

        foreach ($this->dbh->query($query, PDO::FETCH_ASSOC) as $row) {
            $f = array();
            foreach ($row as $k => $v) {
                $k = strtoupper($k);
                if ($k === 'EVT_CONT') {
                    $f[$k] = trim($v);
                } else {
                    $f[$k] = (int)$v;
                }
            }
            $events[] = $f;
        }
        $this->checkError();

        $__EVENTS_QUERY_INFO['query'] = $query;
        $__EVENTS_QUERY_INFO['time_slices_mode'] = $time_slices_mode;
        $__EVENTS_QUERY_INFO['timebegin'] = $timebegin;
        $__EVENTS_QUERY_INFO['timeend'] = $timeend;
        $__EVENTS_QUERY_INFO['weekdays'] = $weekdays;
        @$__EVENTS_QUERY_INFO['time_in_day_begin'] = $time_in_day_begin;
        @$__EVENTS_QUERY_INFO['time_in_day_end'] = $time_in_day_end;

        return $events;
    }

    /**
     *  Добавление параметров камеры
     * @param string $bind_mac 'local'
     * @param int $cam_nr номер камеры
     * @param string $parname название параметра
     * @param string $parval значение параметра
     * @param string $host для какого хоста
     * @param string $user для какого пользователя
     */
    public function addCamera($bind_mac, $cam_nr, $parname, $parval, $host, $user)
    {
        $parval = $parval == null ? 'NULL' : "'$parval'";
        $query = 'INSERT INTO CAMERAS ';
        $query .= '(BIND_MAC, CAM_NR, PARNAME, PARVAL, CHANGE_HOST, CHANGE_USER)';
        $query .= " VALUES ('" . $bind_mac . "'," . $cam_nr . ",'" . $parname . "'," . $parval . ",'" .
            $host . "','" . $user . "')";
        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
    }

    /**
     *
     * Обновление параметров камеры
     *
     * @param string $bind_mac 'local'
     * @param int $cam_nr номер камеры
     * @param string $parname название параметра
     * @param string $parval значение параметра
     * @param string $host для какого хоста
     * @param string $user для какого пользователя
     */
    public function updateCamera($bind_mac, $cam_nr, $parname, $parval, $host, $user)
    {
        $parval = $parval == null ? 'NULL' : "'$parval'";
        $query = 'UPDATE CAMERAS SET';
        $query .= " PARVAL = $parval";
        $query .= ", CHANGE_HOST = '$host'";
        $query .= ", CHANGE_USER = '$user'";
        $query .= ", CHANGE_TIME = NOW()";
        $query .= " WHERE BIND_MAC = '$bind_mac'";
        $query .= " AND CAM_NR = $cam_nr";
        $query .= " AND PARNAME = '$parname'";
        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
    }

    /**
     *  Добавление или изменение параметров камеры
     * @param string $bind_mac 'local'
     * @param int $cam_nr номер камеры
     * @param string $parname название параметра
     * @param string $parval значение параметра
     * @param string $host для какого хоста
     * @param string $user для какого пользователя
     */
    public function replaceCamera($bind_mac, $cam_nr, $parname, $parval, $host, $user)
    {
        $stmt = $this->dbh->prepare(
            'SELECT count(*) FROM CAMERAS WHERE BIND_MAC=? AND CAM_NR=? AND PARNAME=?'
        );
        $stmt->execute([$bind_mac, $cam_nr, $parname]);
        $this->checkError(true, $stmt);
        $count = $stmt->fetchColumn();
        $stmt->closeCursor();

        if (!$count) {
            $this->addCamera($bind_mac, $cam_nr, $parname, $parval, $host, $user);
        } else {
            $this->updateCamera($bind_mac, $cam_nr, $parname, $parval, $host, $user);
        }
    }

    /**
     *
     * Метод получает настройки камеры по умолчанию
     * @param int $cam_nr номер камеры
     * @param string $bind_mac 'local'
     * @return array параметры камеры
     */
    public function getDefCamParams($cam_nr = 0, $bind_mac = 'local')
    {
        $cams = array();

        $query = 'SELECT CAM_NR, PARNAME, PARVAL, CHANGE_HOST, CHANGE_USER, CHANGE_TIME FROM CAMERAS';
        $query .= ' WHERE BIND_MAC=\'' . $bind_mac . '\' AND (CAM_NR=0 OR CAM_NR=' . $cam_nr . ')';

        $stmt = $this->dbh->prepare(
            'SELECT CAM_NR, PARNAME, PARVAL, CHANGE_HOST, CHANGE_USER, CHANGE_TIME FROM CAMERAS ' .
            'WHERE BIND_MAC=? AND (CAM_NR=0 OR CAM_NR=?)'

        );
        if (false === $stmt->execute([$bind_mac, $cam_nr])) {
            $this->checkError(true, $stmt);
        }

        while ($row = $stmt->fetch(PDO::FETCH_ASSOC)) {
            $cams[] = array(
                'CAM_NR' => trim($row[$this->key('CAM_NR')]),
                'PARAM' => trim($row[$this->key('PARNAME')]),
                'VALUE' => trim($row[$this->key('PARVAL')]),
                'CHANGE_HOST' => trim($row[$this->key('CHANGE_HOST')]),
                'CHANGE_USER' => trim($row[$this->key('CHANGE_USER')]),
                'CHANGE_TIME' => trim($row[$this->key('CHANGE_TIME')]),
            );
        }
        $stmt->closeCursor();

        return $cams;
    }

    /**
     *
     * Метод позволяет получить параметры камер
     * @param string $cams_list список камер
     * @param string $param_list список параметров
     * @param string $bind_mac 'local'
     * @return array параметры камер
     */
    public function getCamParams($cams_list = '', $param_list = '', $bind_mac = 'local')
    {
        $cams = array();
        $query = 'SELECT CAM_NR, PARNAME, PARVAL FROM CAMERAS';
        $query .= ' WHERE BIND_MAC=\'' . $bind_mac . '\'';
        if (!empty($cams_list)) {
            $query .= ' AND (CAM_NR=0  OR CAM_NR in(' . $cams_list . '))';
        }
        $query .= ' AND PARNAME IN (' . $param_list . ')'; // AND  PARVAL<>\'\' AND PARVAL IS NOT NULL ';
        $query .= ' ORDER BY CAM_NR';

        foreach ($this->dbh->query($query) as $row) {
            $cams[] = array(
                'CAM_NR' => trim($row[$this->key('CAM_NR')]),
                'PARAM' => trim($row[$this->key('PARNAME')]),
                'VALUE' => trim($row[$this->key('PARVAL')]),
            );
        }

        $this->checkError();

        return $cams;
    }

    /**
     *
     * Метод позволяет получить последний номер камеры
     * @param string $bind_mac 'local'
     * @return int номер камеры
     */
    public function maxCamNr($bind_mac = 'local')
    {
        $query = 'SELECT MAX(CAM_NR) AS LAST_NUM FROM CAMERAS WHERE BIND_MAC = :mac';

        $stmt = $this->dbh->prepare($query, array(PDO::ATTR_CURSOR => PDO::CURSOR_FWDONLY));
        if (!$stmt->execute(array(':mac' => $bind_mac))) {
            $this->checkError(true, $stmt);
        }

        $row = $stmt->fetch(PDO::FETCH_NUM);
        if ($row === false) {
            $this->checkError(true,  $stmt);
        }
        $stmt->closeCursor();

        return isset($row[0]) ? $row[0] : false;
    }

    /**
     *
     * Метод позволяет удалить камеру
     * @param int $cam_nr номер камеры
     * @param string $bind_mac 'local'
     */

    public function deleteCamera($cam_nr, $bind_mac = 'local')
    {
        $query = sprintf('DELETE FROM CAMERAS WHERE BIND_MAC=\'' . $bind_mac . '\' AND CAM_NR=%d', $cam_nr);
        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
    }

    /**
     *
     * Метод позволяет добавить раскладку
     * @param string $display
     * @param int $mon_nr
     * @param string $mon_type
     * @param string $mon_name
     * @param string $remote_addr
     * @param string $login_user
     * @param array $fWINS
     * @param array $vWINS
     * @param string $bind_mac
     */
    public function addLayouts(
        $display,
        $mon_nr,
        $mon_type,
        $mon_name,
        $remote_addr,
        $login_user,
        $fWINS,
        $vWINS,
        $bind_mac = 'local'
    ) {
        $query = sprintf(
            'INSERT INTO LOCAL_LAYOUTS (BIND_MAC, DISPLAY, MON_NR, MON_TYPE, MON_NAME, %s, CHANGE_HOST, CHANGE_USER) ' .
            'VALUES (\'local\', \'%s\', %d, \'%s\', \'%s\', %s, \'%s\', \'%s\')',
            implode(', ', $fWINS),
            $display,
            $mon_nr,
            $mon_type,
            $mon_name,
            implode(', ', $vWINS),
            $remote_addr,
            $login_user
        );
        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
    }

    /**
     *
     * Метод позволяет добавить раскладку для WEB
     * @param int $mon_nr
     * @param string $mon_type
     * @param string $mon_name
     * @param string $remote_addr
     * @param string $login_user
     * @param $PrintCamNames
     * @param $AspectRatio
     * @param $ReconnectTimeout
     * @param $StopHiddenCams
     * @param $allWINS
     * @param string $bind_mac
     * @internal param array $fWINS
     * @internal param array $vWINS
     */
    public function webAddLayouts(
        $mon_nr,
        $mon_type,
        $mon_name,
        $remote_addr,
        $login_user,
        $PrintCamNames,
        $AspectRatio,
        $ReconnectTimeout,
        $StopHiddenCams,
        $allWINS,
        $bind_mac = 'local'
    ) {
        $mon_type = trim($mon_type);
        $mon_name = trim($mon_name);
        $remote_addr = trim($remote_addr);
        $login_user = trim($login_user);
        $AspectRatio = trim($AspectRatio);
        $allWINS = trim($allWINS);
        $bind_mac = trim($bind_mac);

        $query = sprintf(
            'INSERT INTO WEB_LAYOUTS (BIND_MAC, MON_NR, MON_TYPE, SHORT_NAME, PRINT_CAM_NAME , PROPORTION, '
            . 'RECONNECT_TOUT, STOP_HIDDEN_CAMS, WINS, CHANGE_HOST, CHANGE_USER) '
            . 'VALUES (\'local\', %d, \'%s\', \'%s\', %s, \'%s\', %d, %d, \'%s\', \'%s\', \'%s\')',
            $mon_nr,
            $mon_type,
            $mon_name,
            $PrintCamNames,
            $AspectRatio,
            $ReconnectTimeout,
            $StopHiddenCams,
            $allWINS,
            $remote_addr,
            $login_user
        );
        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
    }

    /**
     *
     * Метод позволяет удалить раскладку
     * @param string $display
     * @param int $mon_nr
     * @param string $bind_mac
     */
    public function deleteLayouts($display, $mon_nr, $bind_mac = 'local')
    {
        $query = 'DELETE FROM LOCAL_LAYOUTS';
        $query .= " WHERE BIND_MAC ='$bind_mac'";
        $query .= " AND DISPLAY ='$display'";
        $query .= " AND MON_NR = $mon_nr";
        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
    }

    /**
     *
     * Метод позволяет удалить раскладку для WEB
     * @param int $mon_nr
     * @param string $bind_mac
     */
    public function webDeleteLayouts($mon_nr, $bind_mac = 'local')
    {
        $query = 'DELETE FROM WEB_LAYOUTS';
        $query .= " WHERE BIND_MAC ='$bind_mac'";
        $query .= " AND MON_NR = $mon_nr";
        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
    }

    /**
     *
     * Метод позволяет обновить данные раскладки в БД
     * @param string $display
     * @param unknown_type $mon_nr
     * @param string $mon_type
     * @param string $mon_name
     * @param string $host
     * @param string $user
     * @param array $fWINS
     * @param array $vWINS
     * @param string $bind_mac
     */
    public function updateLayouts(
        $display,
        $mon_nr,
        $mon_type,
        $mon_name,
        $host,
        $user,
        $fWINS,
        $vWINS,
        $bind_mac = 'local'
    ) {
        $query = 'UPDATE LOCAL_LAYOUTS SET ';
        $query .= "MON_TYPE = '$mon_type'";
        $query .= ", MON_NAME = '$mon_name'";
        $query .= ", CHANGE_HOST = '$host'";
        $query .= ", CHANGE_USER = '$user'";

        for ($i = 0; $i < count($vWINS); $i++) {
            if (!empty($vWINS[$i]) /* неважно что '0' даст true, 0-вой камеру тут не будет */) {
                $query .= ", {$fWINS[$i]} = {$vWINS[$i]}";
            } else {
                $query .= ", {$fWINS[$i]} = NULL";
            }
        }

        $query .= " WHERE BIND_MAC ='$bind_mac'";
        $query .= " AND DISPLAY ='$display'";
        $query .= " AND MON_NR = $mon_nr";

        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
    }

    /**
     *
     * Метод позволяет обновить данные раскладки для WEB  в БД
     * @param unknown_type $mon_nr
     * @param string $mon_type
     * @param string $mon_name
     * @param string $host
     * @param string $user
     * @param $PrintCamNames
     * @param $AspectRatio
     * @param $ReconnectTimeout
     * @param $StopHiddenCams
     * @param $allWINS
     * @param string $bind_mac
     * @internal param string $display
     * @internal param array $fWINS
     * @internal param array $vWINS
     */
    public function webUpdateLayouts(
        $mon_nr,
        $mon_type,
        $mon_name,
        $host,
        $user,
        $PrintCamNames,
        $AspectRatio,
        $ReconnectTimeout,
        $StopHiddenCams,
        $allWINS,
        $bind_mac = 'local'
    ) {
        $query = 'UPDATE WEB_LAYOUTS SET ';
        $query .= "MON_TYPE = '$mon_type'";
        $query .= ", SHORT_NAME = '$mon_name'";
        $query .= ", CHANGE_HOST = '$host'";
        $query .= ", CHANGE_USER = '$user'";
        $query .= ", PRINT_CAM_NAME = '$PrintCamNames'";
        $query .= ", PROPORTION = '$AspectRatio'";
        $query .= ", RECONNECT_TOUT = $ReconnectTimeout";
        $query .= ", STOP_HIDDEN_CAMS = $StopHiddenCams";
        $query .= ", WINS = '$allWINS'";
        $query .= " WHERE BIND_MAC ='$bind_mac'";
        $query .= " AND MON_NR = $mon_nr";

        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
    }

    /**
     *
     * Метод позволяет установить раскладку по умолчанию для WEB
     * @param unknown_type $mon_nr - номер раскладки, устанавливаемый по умолчанию
     */
    public function webSetDefLayout($mon_nr)
    {
        $query = 'UPDATE WEB_LAYOUTS SET ';
        $query .= "IS_DEFAULT = 0";
        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }

        $query = 'UPDATE WEB_LAYOUTS SET ';
        $query .= "IS_DEFAULT = 1";
        $query .= " WHERE MON_NR = $mon_nr";

        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
    }

    /**
     *
     * Метод добавляет или обновляет параметры раскладки
     * @param string $display
     * @param unknown_type $mon_nr
     * @param string $mon_type
     * @param string $mon_name
     * @param string $host
     * @param string $user
     * @param array $fWINS
     * @param array $vWINS
     * @param string $bind_mac
     */
    public function replaceLayouts(
        $display,
        $mon_nr,
        $mon_type,
        $mon_name,
        $host,
        $user,
        $fWINS,
        $vWINS,
        $bind_mac = 'local'
    ) {
        $query = 'SELECT count(*) FROM LOCAL_LAYOUTS WHERE BIND_MAC=? AND MON_NR=? AND DISPLAY=?';

        $stmt = $this->dbh->prepare($query);
        if ($stmt === false) {
            $this->checkError();
        }

        if (false === $stmt->execute([$bind_mac, $mon_nr, $display])) {
            $this->checkError(true, $stmt);
        }

        $count = $stmt->fetchColumn();
        $this->checkError(true, $stmt);
        $stmt->closeCursor();

        if (!$count) {
            $this->addLayouts($display, $mon_nr, $mon_type, $mon_name, $host, $user, $fWINS, $vWINS);
        } else {
            $this->updateLayouts($display, $mon_nr, $mon_type, $mon_name, $host, $user, $fWINS, $vWINS);
        }
    }

    /**
     *
     * Метод добавляет или обновляет параметры раскладки для WEB
     * @param unknown_type $mon_nr
     * @param string $mon_type
     * @param string $mon_name
     * @param string $host
     * @param string $user
     * @param $PrintCamNames
     * @param $AspectRatio
     * @param $ReconnectTimeout
     * @param $StopHiddenCams
     * @param $allWINS
     * @param string $bind_mac
     * @internal param string $display
     * @internal param array $fWINS
     * @internal param array $vWINS
     */
    public function webReplaceLayouts(
        $mon_nr,
        $mon_type,
        $mon_name,
        $host,
        $user,
        $PrintCamNames,
        $AspectRatio,
        $ReconnectTimeout,
        $StopHiddenCams,
        $allWINS,
        $bind_mac = 'local'
    ) {
        $mon_type = trim($mon_type);
        $mon_name = trim($mon_name);
        $host = trim($host);
        $user = trim($user);
        $AspectRatio = trim($AspectRatio);
        $allWINS = trim($allWINS);
        $bind_mac = trim($bind_mac);

        $query = 'SELECT count(*) FROM WEB_LAYOUTS WHERE BIND_MAC=? AND MON_NR=?';
        $stmt = $this->dbh->prepare($query);
        if ($stmt === false) {
            $this->checkError();
        }

        if (false === $stmt->execute([$bind_mac, $mon_nr])) {
            $this->checkError(true, $stmt);
        }

        $count = $stmt->fetchColumn();
        $this->checkError(true, $stmt);
        $stmt->closeCursor();

        if (!$count) {
            $this->webAddLayouts(
                $mon_nr,
                $mon_type,
                $mon_name,
                $host,
                $user,
                $PrintCamNames,
                $AspectRatio,
                $ReconnectTimeout,
                $StopHiddenCams,
                $allWINS
            );
        } else {
            $this->webUpdateLayouts(
                $mon_nr,
                $mon_type,
                $mon_name,
                $host,
                $user,
                $PrintCamNames,
                $AspectRatio,
                $ReconnectTimeout,
                $StopHiddenCams,
                $allWINS
            );
        }
    }

    /**
     *
     * Метод позволяет получить параметры раскладки
     * @param string $display
     * @param int $mon_nr
     * @param string $bind_mac
     * @return array параметры
     */
    public function getMonitor($display, $mon_nr, $bind_mac = 'local')
    {
        $stmt = $this->dbh->prepare(
            'SELECT MON_NR, MON_TYPE, MON_NAME, IS_DEFAULT,
            WIN1, WIN2, WIN3, WIN4, WIN5, WIN6, WIN7, WIN8, WIN9,
            WIN10, WIN11, WIN12, WIN13, WIN14, WIN15, WIN16,
            WIN17, WIN18, WIN19, WIN20, WIN21, WIN22, WIN23,
            WIN24, WIN25, WIN26, WIN27, WIN28, WIN29, WIN30,
            WIN31, WIN32, WIN33, WIN34, WIN35, WIN36, WIN37,
            WIN38, WIN39, WIN40,
            WIN41, WIN42, WIN43, WIN44, WIN45, WIN46, WIN47, WIN48, WIN49, WIN50,
            WIN51, WIN52, WIN53, WIN54, WIN55, WIN56, WIN57, WIN58, WIN59, WIN60
            CHANGE_HOST, CHANGE_USER, CHANGE_TIME
            FROM LOCAL_LAYOUTS
            WHERE BIND_MAC=? AND DISPLAY=? AND MON_NR=?'
        );

        if ($stmt === false) {
            $this->checkError();
        }

        if (false === $stmt->execute([$bind_mac, $display, $mon_nr])) {
            $this->checkError(true, $stmt);
        }
        $row = $stmt->fetch();
        if ($row === false) {
            $this->checkError(true, $stmt);
        }
        $stmt->closeCursor();

        return $row;
    }

    /**
     *
     * Метод позволяет получить параметры раскладки для WEB
     * @param int $mon_nr
     * @param string $bind_mac
     * @return array параметры
     */
    public function webGetMonitor($mon_nr, $bind_mac = 'local')
    {

        $stmt = $this->dbh->prepare(
            'SELECT MON_NR, MON_TYPE, SHORT_NAME, IS_DEFAULT, WINS,'
            . ' CHANGE_HOST, CHANGE_USER, CHANGE_TIME, PRINT_CAM_NAME, PROPORTION,'
            . ' RECONNECT_TOUT, STOP_HIDDEN_CAMS'
            . ' FROM WEB_LAYOUTS'
            . ' WHERE BIND_MAC=? AND MON_NR=?'
        );
        if ($stmt === false) {
            $this->checkError();
        }
        if (false === $stmt->execute([$bind_mac, $mon_nr])) {
            $this->checkError(true, $stmt);
        }
        $row = $stmt->fetch();
        if ($stmt === false) {
            $this->checkError();
        }
        $stmt->closeCursor();

        return $row;
    }

    /**
     *
     * Метод позволяет получить параметры всех раскладок
     * @param string $bind_mac
     * @return array раскладки
     */
    public function getLayouts($bind_mac = 'local')
    {
        $mon = array();

        $stmt = $this->dbh->prepare(
            'SELECT * FROM LOCAL_LAYOUTS WHERE BIND_MAC=? ORDER BY MON_NR'
        );
        if ($stmt === false) {
            $this->checkError();
        }

        if (false === $stmt->execute([$bind_mac])) {
            $this->checkError(true, $stmt);
        }

        while ($row = $stmt->fetch()) {
            $m = array();
            foreach ($row as $k => $v) {
                $k = strtoupper($k);
                $m[$k] = trim($v);
            }
            $mon[] = $m;
        }
        if ($row === false) {
            $this->checkError(true, $stmt);
        }

        return $mon;
    }

    /**
     *
     * Метод позволяет получить параметры всех раскладок для WEB
     * или WEB-раскладок, разрешенных для пользователя
     * @param null $user
     * @param string $bind_mac
     * @return array раскладки
     */
    public function webGetLayouts($user = null, $bind_mac = 'local')
    {

        $mon = array();
        $query = 'SELECT * FROM WEB_LAYOUTS';

        $allowed_layouts = array();
        //номер раскладки, указанный в пользовательских настройках первым, устанавливается по умолчанию
        $def_num = null;

        //Если пользователь указан - формируем запрос о разрешенных раскладках
        if (!is_null($user)) {
            $stmt = $this->dbh->prepare( 'SELECT ALLOW_LAYOUTS FROM USERS WHERE USER_LOGIN = ?');
            if ($stmt === false) {
                $this->checkError();
            }

            if (false === $stmt->execute([$user])) {
                $this->checkError(true, $stmt);
            }

            //Определяем разрешенные раскладки
            while ($row = $stmt->fetch()) {
                $l = array();
                foreach ($row as $k => $v) {
                    $k = strtoupper($k);
                    $l[$k] = trim($v);
                }
                $lo[] = $l;
            }
            if ($row === false) {
                $this->checkError(true, $stmt);
            }
            $stmt->closeCursor();

            $allowed_layouts = explode(',', trim($lo[0]["ALLOW_LAYOUTS"]));

            //Первая указанная раскладка используется по умолчанию
            $def_num = $allowed_layouts[0];

            /*если разрешены все раскладки(пустое поле разрешенных раскладок - обнуляем пользователя и раскладку
             по умолчанию)*/
            if (trim($lo[0]["ALLOW_LAYOUTS"]) == '' || $allowed_layouts[0] == '') {
                $user = null;
                $def_num = null;
            } else {
                //определяем перечень разрешенных раскладок пользователя и формируем соотв. запрос
                $lts = "'";
                $lts = $lts . implode("', '", $allowed_layouts) . "'";

                $query .= " WHERE BIND_MAC='$bind_mac' AND MON_NR IN ($lts)";
                $query .= ' ORDER BY MON_NR';
            }
        }

        //Если пользователь не указан или не заданны конкретные раскладки - выбираем все раскладки
        if (is_null($user)) {
            $query .= " WHERE BIND_MAC='$bind_mac'";
            $query .= ' ORDER BY MON_NR';
        }

        foreach ($this->dbh->query($query) as $row) {
            $m = array();
            foreach ($row as $k => $v) {
                $k = strtoupper($k);
                $m[$k] = trim($v);
            }
            $mon[] = $m;
        }
        if ($row === false) {
            $this->checkError(true, $stmt);
        }

        //если в пользовательских настройках указана раскладка по умолчанию
        if ($def_num != null && $user != null) {
            foreach ($mon as $key => $val) {
                $mon[$key]["IS_DEFAULT"] = "0";
                if ($mon[$key]['MON_NR'] == $def_num) {
                    $mon[$key]["IS_DEFAULT"] = '1';
                }
            }
        }
        return $mon;
    }

    /**
     * Метод добавляет пользователя
     *
     * @param string $u_host Допустимые IP-адреса пользователького хоста.
     * @param string $u_name Логин
     * @param string $passwd пароль
     * @param string $groups група
     * @param unknown_type $guest - гостевой доступ
     * @param unknown_type $pda - доступ к PDA версии
     * @param string $u_devacl Доступные камеры
     * @param $u_layouts
     * @param string $u_forced_saving_limit Максимальная длительность принудительной записи (по команде) в минутах
     * @param string $sessions_per_user Ограничение количества одновременных просмотров (камер) пользователем
     * @param string $limit_fps limit_fps, кадров в секунду, [1-25] или sec/frames
     * @param string $nonmotion_fps nonmotion_fps, примеры допустимых значений: "1" - 1 кадр в 1 сек.; "2/1"
     *          - 1 кадр каждые 2 секунды.
     * @param string $limit_kbps limit_kbps, Kбит/сек
     * @param string $session_time session_time - по времени, в минутах
     * @param string $session_volume session_volume - по "закаченному" объёму, в МегаБайтах (10242)
     * @param string $u_longname ФИО
     * @param string $remote_addr хост на котором добавляют
     * @param string $login_user пользователь, который добавляет
     * @return bool результат добавления
     */
    public function addUser(
        $u_host,
        $u_name,
        $passwd,
        $groups,
        $guest,
        $pda,
        $u_devacl,
        $u_layouts,
        $u_forced_saving_limit,
        $sessions_per_user,
        $limit_fps,
        $nonmotion_fps,
        $limit_kbps,
        $session_time,
        $session_volume,
        $u_longname,
        $remote_addr,
        $login_user
    ) {
        $query = sprintf(
            'INSERT INTO USERS
                     ( ALLOW_FROM, USER_LOGIN, PASSWD, STATUS, GUEST, PDA, ALLOW_CAMS, ALLOW_LAYOUTS,
                     MAX_FORCED_REC_MINUTES, MAX_MEDIA_SESSIONS_NB,
                     MAX_VIDEO_FPS, MAX_VIDEO_NONMOTION_FPS, MAX_MEDIA_SESSION_RATE_KB,
                     MAX_MEDIA_SESSION_MINUTES, MAX_MEDIA_SESSION_VOLUME_MB,
                     LONGNAME, CHANGE_HOST, CHANGE_USER, CHANGE_TIME)
                     VALUES ( %s, %s, %s, %u, %b, %b, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, NOW())',
            sql_format_str_val($u_host),
            sql_format_str_val($u_name),
            $this->crypt($passwd),
            $groups,
            $guest,
            $pda,
            sql_format_str_val($u_devacl),
            sql_format_str_val($u_layouts),
            sql_format_int_val($u_forced_saving_limit),
            sql_format_int_val($sessions_per_user),
            sql_format_str_val($limit_fps),
            sql_format_str_val($nonmotion_fps),
            sql_format_int_val($limit_kbps),
            sql_format_int_val($session_time),
            sql_format_int_val($session_volume),
            sql_format_str_val($u_longname),
            sql_format_str_val($remote_addr),
            sql_format_str_val($login_user)
        );

        $res = $this->dbh->exec($query);
        return !$this->checkError(false);
    }

    /**
     *
     * Метод позволяет обновить информацию о пользователе
     * @param string $u_host Допустимые IP-адреса пользователького хоста.
     * @param string $u_name Логин
     * @param string $passwd пароль
     * @param string $groups група
     * @param unknown_type $guest - гостевой доступ
     * @param unknown_type $pda - доступ к PDA версии
     * @param string $u_devacl Доступные камеры
     * @param $u_layouts
     * @param string $u_forced_saving_limit Максимальная длительность принудительной записи (по команде) в минутах
     * @param string $sessions_per_user Ограничение количества одновременных просмотров камер пользователем
     * @param string $limit_fps limit_fps, кадров в секунду, [1-25] или sec/frames
     * @param string $nonmotion_fps nonmotion_fps, примеры допустимых значений: "1" - 1 кадр в 1 сек.; "2/1"
     *          - 1 кадр каждые 2 секунды.
     * @param string $limit_kbps limit_kbps, Kбит/сек
     * @param string $session_time session_time - по времени, в минутах
     * @param string $session_volume session_volume - по "закаченному" объёму, в МегаБайтах (10242)
     * @param string $u_longname ФИО
     * @param string $remote_addr хост на котором добавляют
     * @param string $login_user пользователь, который добавляет
     * @param string $old_u_host старый хост
     * @param string $old_u_name старый логин
     * @return bool результат обновления
     */

    public function updateUser(
        $u_host,
        $u_name,
        $passwd,
        $groups,
        $guest,
        $pda,
        $u_devacl,
        $u_layouts,
        $u_forced_saving_limit,
        $sessions_per_user,
        $limit_fps,
        $nonmotion_fps,
        $limit_kbps,
        $session_time,
        $session_volume,
        $u_longname,
        $remote_addr,
        $login_user,
        $old_u_host,
        $old_u_name
    ) {
        if (empty($passwd) || $passwd[0] !== '*') {
            $pass_upd = ' PASSWD=' . $this->crypt($passwd) . ',';
        } else {
            $pass_upd = '';
        }
        $str = 'UPDATE USERS SET ALLOW_FROM=%s, USER_LOGIN=%s,%s STATUS=%d, GUEST=%b, PDA=%b, ' .
            'ALLOW_CAMS=%s, ALLOW_LAYOUTS=%s ,MAX_FORCED_REC_MINUTES=%s, MAX_MEDIA_SESSIONS_NB=%s, MAX_VIDEO_FPS=%s, ' .
            'MAX_VIDEO_NONMOTION_FPS=%s, MAX_MEDIA_SESSION_RATE_KB=%s, MAX_MEDIA_SESSION_MINUTES=%s, ' .
            'MAX_MEDIA_SESSION_VOLUME_MB=%s, LONGNAME=%s, CHANGE_HOST=%s, CHANGE_USER=%s, ' .
            'CHANGE_TIME=NOW() WHERE ALLOW_FROM=%s AND USER_LOGIN=%s';
        $query = sprintf(
            $str,
            sql_format_str_val($u_host),
            sql_format_str_val($u_name),
            $pass_upd,
            $groups,
            $guest,
            $pda,
            sql_format_str_val($u_devacl),
            sql_format_str_val($u_layouts),
            sql_format_int_val($u_forced_saving_limit),
            sql_format_int_val($sessions_per_user),
            sql_format_str_val($limit_fps),
            sql_format_str_val($nonmotion_fps),
            sql_format_int_val($limit_kbps),
            sql_format_int_val($session_time),
            sql_format_int_val($session_volume),
            sql_format_str_val($u_longname),
            sql_format_str_val($remote_addr),
            sql_format_str_val($login_user),
            sql_format_str_val($old_u_host),
            sql_format_str_val($old_u_name)
        );
        $res = $this->dbh->exec($query);
        return !$this->checkError(false);
    }

    /**
     *
     * Метод позволяет удалить пользователя
     * @param string $u_name логин
     * @param string $u_host хост
     * @param int $u_status статус
     */
    public function deleteUser($u_name, $u_host, $u_status)
    {
        $query = sprintf(
            'DELETE FROM USERS WHERE USER_LOGIN=%s AND ALLOW_FROM=%s AND STATUS=%u',
            sql_format_str_val($u_name),
            sql_format_str_val($u_host),
            $u_status
        );

        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
    }

    /**
     *
     * Метод позволяет получить пароль пользователя
     * @param string $u_name логин
     * @param string $hosts хост
     * @return string пароль
     */
    public function getUserPassword($u_name, $hosts)
    {
        $query = sprintf(
            "SELECT PASSWD FROM USERS WHERE ALLOW_FROM in(%s) AND USER_LOGIN='%s'",
            "'" . implode("','", $hosts) . "'",
            $u_name
        );
        $row = $this->dbh->query($query)->fetch(PDO::FETCH_NUM);
        if ($row === false) {
            $this->checkError();
        }
        return isset($row[0]) ? trim($row[0]) : false;
    }

    /**
     *
     * Метод позволяет обновить пароль пользователя
     * @param string $u_name логин
     * @param string $u_pass пароль
     * @param string $hosts хост
     * @return bool
     */
    public function updateUserPassword($u_name, $u_pass, $hosts)
    {
        $query = sprintf(
            "UPDATE USERS SET PASSWD=%s	 WHERE ALLOW_FROM in(%s) AND USER_LOGIN='%s'",
            $this->crypt($u_pass),
            "'" . implode("','", $hosts) . "'",
            $u_name
        );
        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
        return true;
    }

    /**
     * Метод позволяет получить пользователей
     *
     * @param bool|int $status статус
     * @return array масив пользователей
     */
    public function getUsers($status = false)
    {
        $users = array();
        $query = 'SELECT ALLOW_FROM, USER_LOGIN, GUEST, PDA, PASSWD, STATUS, ALLOW_CAMS, ALLOW_LAYOUTS, ' .
            'MAX_FORCED_REC_MINUTES,  MAX_MEDIA_SESSIONS_NB, MAX_VIDEO_FPS, MAX_VIDEO_NONMOTION_FPS, ' .
            'MAX_MEDIA_SESSION_RATE_KB, MAX_MEDIA_SESSION_MINUTES, MAX_MEDIA_SESSION_VOLUME_MB, LONGNAME, ' .
            'CHANGE_HOST, CHANGE_USER, CHANGE_TIME FROM USERS';
        if ($status) {
            $query .= ' WHERE STATUS=?';
        }
        $query .= ' ORDER BY ALLOW_FROM, USER_LOGIN';

        $stmt = $this->dbh->prepare($query);
        if ($stmt === false) {
            $this->checkError();
        }

        if (false === $stmt->execute($status ? [$status] : [])) {
            $this->checkError(true, $stmt);
        }

        while ($row = $stmt->fetch()) {
            $users[] = array(
                'HOST' => trim($row[$this->key('ALLOW_FROM')]),
                'USER' => trim($row[$this->key('USER_LOGIN')]),
                'PASSWD' => trim($row[$this->key('PASSWD')]),
                'GUEST' => trim($row[$this->key('GUEST')]),
                'PDA' => trim($row[$this->key('PDA')]),
                'STATUS' => trim($row[$this->key('STATUS')]),
                'ALLOW_CAMS' => trim($row[$this->key('ALLOW_CAMS')]),
                'ALLOW_LAYOUTS' => trim($row[$this->key('ALLOW_LAYOUTS')]),
                'MAX_FORCED_REC_MINUTES' => trim($row[$this->key('MAX_FORCED_REC_MINUTES')]),
                'MAX_MEDIA_SESSIONS_NB' => trim($row[$this->key('MAX_MEDIA_SESSIONS_NB')]),
                'MAX_VIDEO_FPS' => trim($row[$this->key('MAX_VIDEO_FPS')]),
                'MAX_VIDEO_NONMOTION_FPS' => trim($row[$this->key('MAX_VIDEO_NONMOTION_FPS')]),
                'MAX_MEDIA_SESSION_RATE_KB' => trim($row[$this->key('MAX_MEDIA_SESSION_RATE_KB')]),
                'MAX_MEDIA_SESSION_MINUTES' => trim($row[$this->key('MAX_MEDIA_SESSION_MINUTES')]),
                'MAX_MEDIA_SESSION_VOLUME_MB' => trim($row[$this->key('MAX_MEDIA_SESSION_VOLUME_MB')]),
                'LONGNAME' => trim($row[$this->key('LONGNAME')]),
                'CHANGE_HOST' => trim($row[$this->key('CHANGE_HOST')]),
                'CHANGE_USER' => trim($row[$this->key('CHANGE_USER')]),
                'CHANGE_TIME' => trim($row[$this->key('CHANGE_TIME')]),
            );
        }
        if ($row === false) {
            $this->checkError(true, $stmt);
        }

        return $users;
    }

    /***
     * Метод выключить все камеры work=off
     * Используется для автоматизации конфигурирования.
     * @return bool
     */
    public function offAllCameras($bind_mac = 'local')
    {
        $stmt = $this->dbh->prepare(
            "UPDATE CAMERAS SET PARVAL='0' ".
            "WHERE BIND_MAC=? AND CAM_NR>0 AND PARNAME='work' AND PARVAL != '0'"
        );
        if ($stmt === false) {
            $this->checkError();
        }

        if (false === $stmt->execute([$bind_mac])) {
            $this->checkError(true, $stmt);
        }
        return true;
    }

    /***
     * Изменяет дисплей (правый/левый) для локальной раскладки.
     * @return bool
     */
    public function updateLocalLayoutDisplay($display, $layout_nr = false, $bind_mac = 'local')
    {
        $params = [
            ':display' => $display,
            ':bind_mac' => $bind_mac
        ];

        $query = 'UPDATE LOCAL_LAYOUTS SET DISPLAY=:display WHERE BIND_MAC=:bind_mac';

        if ($layout_nr !== false) {
            $query .= ' AND ' . $this->whereIntColumnValue('MON_NR', $layout_nr);
        }

        $stmt = $this->dbh->prepare($query);
        if ($stmt === false) {
            $this->checkError();
        }

        if (false === $stmt->execute($params)) {
            $this->checkError(true, $stmt);
        }
        return true;
    }

    /**
     * Метод позволяет обновить данные раскладки в БД
     * @param string $display
     * @param integer $mon_nr
     * @param string $host
     * @param string $user
     * @param array $wins { "winNr" => camNr, ...}
     * @param string $bind_mac
     */
    public function updateLayoutWins(
        $display_new,
        $display_now,
        $mon_nr,
        $host,
        $user,
        $wins,
        $bind_mac = 'local'
    ) {
        $query = 'UPDATE LOCAL_LAYOUTS SET ';
        $query .= "DISPLAY = '$display_new'";
        $query .= ", CHANGE_HOST = '$host'";
        $query .= ", CHANGE_USER = '$user'";

        foreach ($wins as $winNr => $camNr) {
            if (!empty($camNr) /* неважно что '0' даст true, 0-вой камеру тут не будет */) {
                $query .= ", WIN{$winNr} = {$camNr}";
            } else {
                $query .= ", WIN{$winNr} = NULL";
            }
        }
        $query .= " WHERE BIND_MAC ='$bind_mac'";
        $query .= " AND DISPLAY ='$display_now'";
        $query .= " AND MON_NR = $mon_nr";

        if (false === $this->dbh->exec($query)) {
            $this->checkError();
        }
    }

    private function key($str)
    {
        if ($this->dbAdapter === 'pgsql') {
            return strtolower($str);
        }
        return $str;
    }

    public function dateRange($YMDH, $colName)
    {
        if (is_null($YMDH) || !is_array($YMDH)) {
            throw new \InvalidArgumentException('Unexpected $YMDH type: "' . gettype($YMDH) . '"');
        }
        if (empty($YMDH)) {
            return '';
        }

        if (gettype($colName) !== 'string' || strlen($colName) === 0) {
            throw new \InvalidArgumentException('Unexpected $colName type or value: "' . $colName . '"');
        }

        @list($year, $month, $day, $hour) = $YMDH;
        if (is_null($year)) {
            throw new \InvalidArgumentException('Unexpected $YMDH value: [' . implode(',', $YMDH) . ']');
        }

        $ds = sprintf(
                "%'.04d-%'.02d-%'.02d %'.02d:00:00",
                !is_null($year) ? intval($year) : 0,
                !is_null($month) ? intval($month) : 1,
                !is_null($day) ? intval($day) : 1,
                !is_null($hour) ? intval($hour) : 0);

        $begin = \DateTime::createFromFormat(self::SQLDATETIMEFMT, $ds);

        if ($begin->format(self::SQLDATETIMEFMT) !== $ds) {
            throw new \InvalidArgumentException('Unexpected $YMDH value: [' . implode(',', $YMDH) . ']');
        }

        $daysInMonth = intval($begin->format('t'));

        if (!is_null($hour)) {
            /*
            $y1 = $y2 = !is_null($year) ? intval($year) : 0;
            $m1 = $m2 = !is_null($month) ? intval($month) : 0;
            $d1 = $d2 = !is_null($day) ? intval($day) : 0;
            $h1 = intval($hour);
            if ($h1 < 23) {
                $h2 = $h1 + 1;
            } else {
                $h2 = 0;
                $d2 =
            }
             */
            $modifier = '+1 hour';
        } else if (!is_null($day)) {
            $modifier = '+1 day';
        } else if (!is_null($month)) {
            $modifier = '+1 month';
        } else {
            $modifier = '+1 year';
        }

        $end = clone ($begin);
        $end->modify($modifier);

        return sprintf(
            "AND %s >= '%s' AND %s < '%s'",
            $colName, $begin->format(self::SQLDATETIMEFMT),
            $colName, $end->format(self::SQLDATETIMEFMT)
        );
    }

    protected function datePart($type, $value)
    {
        if ($this->dbAdapter === 'pgsql') {
            switch ($type) {
                case 'year':
                    $str = "date_part('year', %%)";
                    break;
                case 'month':
                    $str = "date_part('month', %%)";
                    break;
                case 'day':
                    $str = "date_part('day', %%)";
                    break;
                case 'hour':
                    $str = "date_part('hour', %%)";
                    break;
                case 'weekday':
                    $str = "date_part('dow', %%)";
                    break;
                case 'time':
                    $str = "%%::time";
                    break;
                case 'timestamp':
                    $str = "extract(epoch from %%::timestamp at time zone 'utc')";
                    break;
            }
        } else {
            switch ($type) {
                case 'year':
                    $str = 'YEAR(%%)';
                    break;
                case 'month':
                    $str = 'MONTH(%%)';
                    break;
                case 'day':
                    $str = 'DAYOFMONTH(%%)';
                    break;
                case 'hour':
                    $str = 'HOUR(%%)';
                    break;
                case 'weekday':
                    $str = 'weekday(%%)';
                    break;
                case 'time':
                    $str = 'time(%%)';
                    break;

                case 'timestamp':
                    $str = "UNIX_TIMESTAMP(CONVERT_TZ(%%, '+00:00', @@global.time_zone))";
                    break;

            }
        }
        $str = str_replace('%%', $value, $str);

        return $str;
    }

    private function dateFormat($value)
    {
        if ($this->dbAdapter === 'pgsql') {
            $str = "to_char(%%, 'yyyy_mm_dd_hh24')";
        } else {
            $str = "DATE_FORMAT(%%, '%Y_%m_%d_%H')";
        }
        $str = str_replace('%%', $value, $str);
        return $str;
    }

    private function timediff($d1, $d2)
    {
        if ($this->dbAdapter === 'pgsql') {
            $str = $d1 . "-" . $d2;
        } else {
            $str = "TIMEDIFF(" . $d1 . " , " . $d2 . ")";
        }
        return $str;
    }

    private function crypt($value)
    {
        if (empty($value)) {
            return "''";
        }

        return "'" . crypt($value) . "'";
    }

    protected function whereIntColumnValue($col_name, $col_value = "")
    {
        if (@empty($col_name)) {
            throw new InvalidArgumentException("empty column name");
        }
        if (@empty($col_value)) {
            throw new InvalidArgumentException("empty '$col_name' column value");
        }
        switch (gettype($col_value)) {
            case 'array':
                return "$col_name in (" . implode(",", $col_value) . ")";
            case "string":
                if (false !== strpos($col_value, ",")) {
                    return "$col_name in (" . $col_value  . ")";
                }
                // else (non CSV) - no break, use as signle number
            case "integer":
                return "$col_name = $col_value";
            default:
                throw new InvalidArgumentException("invalid '$col_name' value");
        }
    }
}
