<?php
/**
 * Copyright (C) 2004-2021 "Setevye informacionnye sistemy", Ltd (Russia)
 *                          en: "Network Information Systems"
 */

/**
 * @file lib/config.inc.php
 * @brief Файл обеспечивает инициализацию глобальных переменных,<br />а также содержит функции для установки
 * параметров конфигурации
 *
 * Файл реализует функции:
 * <ul>
 *    <li> Инициализации глобальных переменных </li>
 *    <li> Загрузки конфигурационных настроек из файлов конфигурации /etc/avreg/site-defaults.php </li>
 *    <li> Авторизации пользователя </li>
 *    <li> Настройки параметров камер </li>
 * </ul>
 *
 *
 */

define("OPT_TYPE_BOOL", 1);
define("OPT_TYPE_INT", 2);
define("OPT_TYPE_INTPROC", 3); // int or %
define("OPT_TYPE_CHECK", 4);
define("OPT_TYPE_STRING", 5);
define("OPT_TYPE_STRING200", 6);
define("OPT_TYPE_TEXT", 7);
define("OPT_TYPE_PASSWORD", 8);
define("OPT_TYPE_ARRAY", 9);
define("OPT_TYPE_SET", 10);

if ((!empty($_SERVER['HTTPS']) && $_SERVER['HTTPS'] !== 'off') || $_SERVER['SERVER_PORT'] == 443) {
    $SSL_ACCESS_USED = true;
    if ($_SERVER['SERVER_PORT'] == 443) {
        $SERVER_PORT_STR = '';
    } else {
        $SERVER_PORT_STR = ':' . $_SERVER['SERVER_PORT'];
    }
} else {
    $SSL_ACCESS_USED = false;
    if ($_SERVER['SERVER_PORT'] == 80) {
        $SERVER_PORT_STR = '';
    } else {
        $SERVER_PORT_STR = ':' . $_SERVER['SERVER_PORT'];
    }
}
$_SERVER['REQUEST_SCHEME'] = $SSL_ACCESS_USED ? 'https' : 'http';

$SERVER_SOFT = [
    'name_lo' => null, // lighttpd | apache | nginx
    'ver_str' => null       // "1.2.34" | null
];

if (@!empty($_SERVER['SERVER_SOFTWARE'])) {
    if (preg_match('/^([a-zA-Z]+)\/([0-9\.\-:\+]+)/', $_SERVER['SERVER_SOFTWARE'], $matches)) {
        $SERVER_SOFT['name_lo'] = strtolower($matches[1]);
        $SERVER_SOFT['ver_str'] = $matches[2];
    }
}

require_once('/etc/avreg/site-defaults.php');
require_once(__DIR__ . '/check-file-mode.inc.php');
require_once(__DIR__ . '/confparse.php');

/// Путь к директории сайта
$wwwdir = $conf['site-dir'] . '/';

/*
if (empty($conf['prefix']))
   $wwwdir = $_SERVER['DOCUMENT_ROOT'] . '/';
else
   $wwwdir = $_SERVER['DOCUMENT_ROOT'] . $conf['prefix'] . '/';
 */
require($wwwdir . 'lib/grab_globals.lib.php');

function is_assoc($array) {
  foreach (array_keys($array) as $k => $v) {
    if ($k !== $v)
      return true;
  }
  return false;
}

/**
 * @brief Форматирование и вывод дампа переменной
 * @param unknown_type $var переменная
 *
 */
function tohtml($var)
{
    print '<div class="dump"><pre class="dump">' . "\n";
    var_dump($var);
    print '</pre></div>' . "\n";
}

/**
 * Функция парсит строки чилел или диапазонов в csv формате в масив чисел
 * @param string $str csv строка
 * @return array
 */
function parse_list_of_numbers($str, $options = array())
{
   global $debug;

   $defOptions = array(
      'strictly' => true,
      'sepPattern' => '/[,;]+/',
      'chunksLimit' => -1,
      'uniq' => true,
      'sort' => true,
      'min' => null,
      'max' => null
   );
   $options = array_merge($defOptions, $options);
   $res = array();
   foreach (preg_split($options['sepPattern'], $str, $options['chunksLimit'], PREG_SPLIT_NO_EMPTY) as $value) {
      $value = trim($value);
      if (is_numeric($value)) {
         $res[] = (int)$value;
      } else {
         /* check range */
         if (preg_match('/^\s*([\+-]?\d+)\s*(?:-|\.\.)\s*([\+-]?\d+)\s*$/', $value, $matches)) {
            $start = (int)$matches[1];
            $end = (int)$matches[2];
            if ($start >= $end) {
               $errMsg = 'Invalid range [' . $value . '].';
               if ($options['strictly']) {
                  throw new RangeException($errMsg);
               } else {
                  if ($debug) {
                     error_log($errMsg . ' Skipped.');
                  }
                  continue;
               }
            }
            for (; $start <= $end; $start++) {
               $res[] = (int)$start;
            }
         } else {
            // error
            $errMsg = 'Unexpected int or int range value "' . $value . '".';
            if ($options['strictly']) {
               throw new InvalidArgumentException($errMsg);
            } else {
               if ($debug) {
                  error_log($errMsg . ' Skipped.');
               }
               continue;
            }
         }
      }
   }

   if ($options['uniq']) {
      $res = array_unique($res, SORT_NUMERIC);
   }

   if ($options['sort']) {
      sort($res, SORT_NUMERIC);
   }

   if (!is_null($options['min']) || !is_null($options['max'])) {
      for ($i = 0, $I = count($res); $i < $I; $i += 1) {
         $value = $res[$i];
         if (!is_null($options['min'])) {
            if ($value < $options['min']) {
               $errMsg = 'A value can not be less than ' . $options['min'] . '.';
               if ($options['strictly']) {
                  throw new RangeException($errMsg);
               } else {
                  unset($res[$i]);
                  if ($debug) {
                     error_log($errMsg . ' Skipped.');
                  }
                  continue;
               }
            }
         }
         if (!is_null($options['max'])) {
            if ($value > $options['max']) {
               $errMsg = 'A value can not be greater than ' . $options['max'] . '.';
               if ($options['strictly']) {
                  throw new RangeException('A value can not be greater than ' . $options['max']);
               } else {
                  unset($res[$i]);
                  if ($debug) {
                     error_log($errMsg . ' Skipped.');
                  }
                  continue;
               }
            }
         }
      }
      $res = array_values($res); // XXX reindex after unset to min/max checking
   }

   if (count($res) > 0) {
      return $res;
   } else {
      $errMsg = 'Empty int array parsed from string "' . $str . '".';
      if ($options['strictly']) {
         throw new RangeException($errMsg);
      } else {
         error_log($errMsg);
      }
      return false;
   }
} /* parse_list_of_numbers() */

$to_int_func = function ($value) {
    if (!settype($value, 'int')) {
        die('crack or hack?');
    }
    return $value;
};

function request_int_array($parname, $permissible_array = false, $sort = true)
{
    global $to_int_func;

    $ret = false;

    if (!isset($_REQUEST[$parname]) || empty($_REQUEST[$parname])) {
        return false;
    }

    if (is_array($_REQUEST[$parname]) && count($_REQUEST[$parname]) > 0) {
        $ret = array_map($to_int_func, $_REQUEST[$parname]);
    } else if (is_string($_REQUEST[$parname])) {
        $ret = parse_list_of_numbers($_REQUEST[$parname], array(
            'sort' => false,
            'uniq' => false,
            'strictly' => false
        ));
    } else {
        return false;
    }

    if (is_array($ret) && count($ret) > 0) {
        if ($sort) {
            sort($ret, SORT_NUMERIC);
        }

        $ret = array_unique($ret, SORT_NUMERIC);

        if (is_array($permissible_array) && count($permissible_array) > 0) {
            $a = array_intersect($permissible_array, $ret);
            $ret = array_values($a);
        }
    }

    return $ret;
}


function get_avreg_profiles($_conf)
{
   $reta = array();

   if ($handle = opendir($_conf['profiles-dir'])) {
      while (false !== ($entry = readdir($handle))) {
         if (preg_match('/^[A-Za-z0-9][A-Za-z0-9_\.:\-]+$/', $entry)) {
            $reta[] = $_conf['profiles-dir'] . '/' . $entry;
         }
      }
      closedir($handle);
   }

   if (count($reta) > 0) {
      sort($reta); // , SORT_STRING | SORT_FLAG_CASE);
      return $reta;
   } else {
      return false;
   }
}

/**
 * Функция позволяет получить настройки камер с учётом разбиения на профили
 * @param string $application
 * @return array масив настроек
 */
function load_profiles_cams_confs($applications = 'avreg-site')
{
    if (!empty($GLOBALS['AVREG_PROFILE']) || empty($GLOBALS['EXISTS_PROFILES'])) {
        return false;
    }

    $cams_profiles = array();
    $profiles_conf = array();
    $i = 0;

    foreach ($GLOBALS['EXISTS_PROFILES'] as &$profile) {
        $a = confparse($GLOBALS['conf'], $applications, $profile);
        if (empty($a) || !array_key_exists('devlist', $a) || empty($a['devlist'])) {
            continue;
        }
        $profiles_conf[$i] = $a;
        $b = parse_list_of_numbers($a['devlist']);
        foreach ($b as &$c) {
            $cams_profiles[$c] = & $profiles_conf[$i];
        }
        $i++;
    }
    return $cams_profiles;
} /* get_profiles_cams_confs() */

function get_storage_dirs()
{
    global $conf;

    $storage_dirs = [];

    if (!empty($GLOBALS['EXISTS_PROFILES'])) {
        foreach ($GLOBALS['EXISTS_PROFILES'] as &$profile) {
            $a = confparse($conf, ['avregd', 'avreg-site'], $profile);
            if (empty($a) || !array_key_exists('storage-dir', $a) || empty($a['storage-dir'])) {
                continue;
            }
            $s = $a['storage-dir'];
            $is_new = true;
            foreach ($storage_dirs as $profile_name => $profile_conf) {
                if ($s === $profile_conf['storage-dir']) {
                    $is_new  = false;
                    break;
                }
            }
            if ($is_new) {
                $profile_name = basename($profile);
                $storage_dirs[$profile_name] = [
                    'storage-dir' => $s,
                    'media-alias' => $a['media-alias']
                ];
                // XXX devlist need to merge if no $is_new
            }
        }
    }

    if (empty($storage_dirs)) {
        $profile_name = empty($GLOBALS['AVREG_PROFILE']) ? '_' : $GLOBALS['AVREG_PROFILE'];
        $storage_dirs[$profile_name] = [
            'storage-dir' => $conf['storage-dir'],
            'media-alias' => $conf['media-alias']
        ];
    }
    return $storage_dirs;
}

$sip = & $_SERVER['SERVER_ADDR'];
$named = & $_SERVER['SERVER_NAME'];

///  ip сервера
$localip = ip2long($sip);
///  user agent пользователя
$ua = @strtolower($_SERVER['HTTP_USER_AGENT']);
///  Движок браузера пользователя
$MSIE = $GECKO = $PRESTO = $WEBKIT = $MS_EDGE = $WEBKIT_SAFARI = false;
$IS_ANDROID = $IS_IPAD = $IS_IPHONE = $IS_IPOD = $IS_WINPHONE = false;

if (preg_match('/ (msie|trident)/', $ua)) {
    $MSIE = true;
} elseif (false !== strpos($ua, 'edge/')) {
    $MS_EDGE = true;
} elseif (false !== strpos($ua, 'presto')) {
    $PRESTO = true;
} elseif (false !== strpos($ua, 'webkit') || false !== strpos($ua, 'khtml')) {
    $WEBKIT = true;
    $WEBKIT_SAFARI = (false !== strpos($ua, 'safari') && false === strpos($ua, 'chrome'));
} elseif (false !== strpos($ua, 'gecko')) {
    /* gecko последним т.к. многие добавляют gecko в U-A в дополнение к основному движку */
    $GECKO = true;
}

if (false != strpos($ua, 'android')) {
    $IS_ANDROID = true;
} elseif (false != strpos($ua, 'iphone')) {
    $IS_IPHONE = true;
} elseif (false != strpos($ua, 'ipad')) {
    $IS_IPAD = true;
} elseif (false != strpos($ua, 'ipod')) {
    $IS_IPOD = true;
} elseif (false != strpos($ua, 'windows phone')) {
    $IS_WINPHONE = true;
}
$IS_IOS = $IS_IPHONE || $IS_IPAD || $IS_IPOD;
$IS_MOBILE = $IS_ANDROID || $IS_IOS || $IS_WINPHONE;

///  логин пользователя
$login_user = 'unknown';
///  имя пользователя
$login_user_name = 'unknown';
///  ip пользователя
$remote_addr = 'unknown';
///  host пользователя
$login_host = 'unknown';
///  масив пользователей
$users = array();
///  информация пользователя
$user_info = null;

$link = null;
///  статус пользователя
$user_status = 555;
///  статус пользователя групы install
$install_status = 1;
///  статус пользователя групы admin
$admin_status = 2;
///  статус пользователя групы arch
$arch_status = 3;
///  статус пользователя групы operator
$operator_status = 4;
///  статус пользователя групы view
$viewer_status = 5;
///  являеться ли пользователь в групе install
$install_user = false;
///  являеться ли пользователь в групе admin
$admin_user = false;
///  являеться ли пользователь в групе arch
$arch_user = false;
///  являеться ли пользователь в групе operator
$operator_user = false;
///  являеться ли пользователь в групе view
$viewer_user = false;
///	 Доступность PDA версии
$allow_pda = false;

///  кодировка
$chset = 'UTF-8';
///  язык
$lang = 'ru'; // TODO use Accept-Language header or other request param
///  локаль
$locale_str = 'ru_RU.' . $chset;
mb_internal_encoding('UTF-8');

if ($_SERVER['REMOTE_ADDR'] === '::1') {
    $_SERVER['REMOTE_ADDR'] = '127.0.0.1';
}

if ($_SERVER['REMOTE_ADDR'] === '127.0.0.1') {
    $remote_addr = 'localhost';
} else {
    $remote_addr = & $_SERVER['REMOTE_ADDR'];
}

/* set timezone "You are *required* to use the date.timezone setting or the date_default_timezone_set() function." */
$tzname = @file_get_contents('/etc/timezone'); // debian-style
if (!empty($tzname)) {
    date_default_timezone_set(rtrim($tzname));
}

///  шрифт текста
$font_family = 'sans-serif';
///  размер шрифта
$font_size = 'small';
///  цвет контента
$ContentColor = '#F5F5F5';
///
$inactive_h_color = 'darkGray';
///  цвет ошибки
$warn_color = '#003366';
///  цвет критической ошибки
$error_color = '#CC3333';
///  цвет
$rowHiLight = '#FFFFCC';
///  цвет заголовка
$header_color = '#D0DCE0';
///  цвет
$TextHiLight = '#009900';

///  цвет
$NotSetParColor = '#585858';
///  цвет
$ParDefColor = '#6633FF';
///  цвет
$ParSetColor = 'Red';

///  максимальное количество камер
$MAX_CAM = 1000;
///  ширина левой колонки
$cfg['LeftWidth'] = '170';
///  цвет фона левой колонки
$left_bgcolor = '#D0DCE0';
///  тег таблиц
$tabletag = '<table cellspacing="0" border="1" cellpadding="2" align="center" style="border-collapse:collapse;>';

$patternIP = '[1-9]\d{1,2}\.\d{1,3}\.\d{1,3}\.\d{1,3}';
$patternAllowedIP = '/^(' . $patternIP . '|any|localhost|*)$/';
$patternUser = '/^[A-Za-z0-9_\-]{4,16}$/';
$patternPasswd = '/^[A-Za-z0-9_\-]{0,16}$/';

$WellKnownAspects = array(
    array(4, 3),
    array(10, 8),
    array(11, 9),
    array(3, 2),
    array(16, 9),
    array(16, 10),
    array(5, 4)
);

/**
 *
 * Функция позволяет получить информацию о пользователе
 * @param string $ipacl хосты
 * @param string $name логин
 * @return информация о пользователе или False
 */
function get_user_info($ipacl, $name)
{
    if (!isset($ipacl) || !isset($name)) {
        return false;
    }
    if (empty($ipacl) || empty($name)) {
        return false;
    }

    foreach ($GLOBALS['users'] as $ui) {
        if (0 === strcasecmp($ui['HOST'], $ipacl) &&
            0 === strcmp($ui['USER'], $name)
        ) {
            return $ui;
        }
    }
    return false;
}

/**
 * Функция получает информацию по пользователю по заданым параметрам
 * @param array $addr хосты
 * @param string $mask маска сети
 * @param string $name логин
 * @return информация о пользователе или False
 */
function avreg_find_user($addr, $mask, $name)
{
    if (!isset($addr) || !isset($mask) || !isset($name)) {
        return false;
    }
    if (is_null($addr) || is_null($mask) || is_null($name)) {
        return false;
    }
    $found = false;
    foreach ($GLOBALS['users'] as $ui) {
        if (0 !== strcmp($ui['USER'], $name)) {
            continue;
        }

        $ipacl = avreg_inet_network($ui['HOST']);
        if ($ipacl === false) {
            continue;
        } // FIXME - may be error/warning?

        $found = avreg_ipv4_cmp(
            $addr,
            $mask,
            $ipacl['addr'],
            $ipacl['mask']
        );
        /*
              syslog(LOG_ERR,sprintf("equal = %d, 0x%X/0x%X 0x%X/0x%X",
                                      $found,
                                      $addr, $mask,
                                      $ipacl['addr'],  $ipacl['mask']));
         */
        if ($found !== false) {
            break;
        }
    }

    if ($found && $GLOBALS['conf']['debug']) {
        syslog(
            LOG_ERR,
            sprintf(
                'ACL %s %s@%s %s/%s',
                $found ? "success" : "failed",
                $name,
                long2ip($addr),
                $ipacl['addr_a'],
                $ipacl['mask_a']
            )
        );
    }

    return ($found) ? $ui : false;
}

/**
 *
 *
 * @param string $dev_acl_str
 */
function parse_dev_acl($dev_acl_str = null)
{
    if (empty($dev_acl_str)) {
        return true;
    } /* any */
    $chunks = explode(',', $dev_acl_str);
    $ret = array();
    foreach ($chunks as &$sub) {
        if (preg_match('/^\s*(\d+)\s*$/', $sub, $matches)) {
            $ret[] = (Int)$matches[1];
        } elseif (preg_match('/^\s*(\d+)\s*-\s*(\d+)\s*$/', $sub, $matches)) {
            $_start = (Int)$matches[1];
            $_stop = (Int)$matches[2];
            if ($_start >= $_stop) {
                return false;
            }
            $ret = array_merge($ret, range($_start, $_stop));
        } else {
            return false;
        }
    }

    if (false === sort($ret, SORT_NUMERIC)) {
        return false;
    }
    return array_unique($ret);
}

/**
 *
 * Функция формирует строку размера файла
 * @param int $fsize_KB размер в килобайтах
 * @return string
 */
function filesizeHuman($fsize_KB)
{
    if ($fsize_KB >= 1048576) {
        return sprintf("%0.1f %s", $fsize_KB / 1048576, $GLOBALS['byteUnits'][3]);
    } elseif ($fsize_KB >= 1024) {
        return sprintf("%0.1f %s", $fsize_KB / 1024, $GLOBALS['byteUnits'][2]);
    } elseif ($fsize_KB >= 1) {
        return sprintf("%d %s", $fsize_KB, $GLOBALS['byteUnits'][1]);
    } else {
        return sprintf("%d %s", $fsize_KB * 1024, $GLOBALS['byteUnits'][0]);
    }
}

/**
 *
 * Функция формирует строку времени
 * @param int $sec время в секундах
 * @return string
 */
function DeltaTimeHuman($sec)
{
    settype($sec, 'int');
    if ($sec >= 86400) {
        return sprintf('%.1f дня', $sec / 86400);
    } elseif ($sec >= 3600) {
        return sprintf('%.1f час', $sec / 3600);
    } elseif ($sec >= 60) {
        return sprintf('%.1f мин', $sec / 60);
    } else {
        return sprintf('%d сек', $sec);
    }
}

/**
 *
 * Функция формирует строку времени
 * @param int $sec время в секундах
 * @return string
 */
function ETA($sec)
{
    settype($sec, 'int');
    if ($sec < 0) {
        return '--:--:--';
    } elseif ($sec === 0) {
        return '00:00:00';
    }
    $_sec = $sec % 60;
    $_min = (int)($sec / 60) % 60;
    $_hour = (int)($sec / 3600);
    return sprintf('%02u:%02u:%02u', $_hour, $_min, $_sec);
}

/* $start,$finish - unix timestamp */
/**
 *
 * Функция формирует строку промежутка времени
 * @param int $start дата и время начала в unix timestamp
 * @param int $finish дата и время окончания в unix timestamp
 * @param bool $print_year учитывать год
 * @param bool $print_sec учитывать секунды
 * @return string
 */
function TimeRangeHuman($start, $finish, $print_year = false, $print_sec = false)
{
    if ($print_sec) {
        $time_fmt = '%02u:%02u:%02u';
    } else {
        $time_fmt = '%02u:%02u';
    }

    $start_tm = localtime($start, true);
    $finish_tm = localtime($finish, true);
    $same_year = ($start_tm['tm_year'] === $finish_tm['tm_year']);
    $same_month = $same_year && ($start_tm['tm_mon'] === $finish_tm['tm_mon']);
    $same_day = $same_month && ($start_tm['tm_mday'] === $finish_tm['tm_mday']);

    $const_str = '';
    $diff_str = '';

    if ($print_sec) {
        $start_time = sprintf('%02u:%02u:%02u', $start_tm['tm_hour'], $start_tm['tm_min'], $start_tm['tm_sec']);
        $finish_time = sprintf('%02u:%02u:%02u', $finish_tm['tm_hour'], $finish_tm['tm_min'], $finish_tm['tm_sec']);
    } else {
        $start_time = sprintf('%02u:%02u', $start_tm['tm_hour'], $start_tm['tm_min']);
        $finish_time = sprintf('%02u:%02u', $finish_tm['tm_hour'], $finish_tm['tm_min']);
    }

    if ($finish < $start /* $finish = null or 0 */) {
        $_date = strftime($print_year ? '%Y %b %d(%a)' : '%b %d(%a)', $start);
        return ("[ $_date $start_time - ??? ]");
    }

    if ($same_day) {
        $const_str = strftime($print_year ? '%Y %b %d(%a) ' : '%b %d(%a) ', $start);
        $diff_str = sprintf('%s - %s', $start_time, $finish_time);
    } else {
        if ($same_month) {
            $const_str = strftime($print_year ? '%Y %b ' : '%b ', $start);
            $date_fmt = '%d(%a)';
        } elseif ($same_year) {
            if ($print_year) {
                $const_str = sprintf('%04u ', $start_tm['tm_year'] + 1900);
            }
            $date_fmt = '%b %d(%a)';
        } else {
            $date_fmt = '%Y %b %d(%a)';
        }
        $diff_str = sprintf(
            "%s $start_time - %s $finish_time",
            strftime($date_fmt, $start),
            strftime($date_fmt, $finish)
        );
    }
    return ($const_str . '[ ' . $diff_str . ' ]');
}

/**
 *
 * Функция получает имя камеры
 * @param string $_text_left имя камеры
 * @return string имя камеры
 */
function getCamName($_text_left)
{
    if (empty($_text_left)) {
        return $GLOBALS['strNotTextLeft'];
    } else {
        return $_text_left;
    }
}

function shortCamName($camNr, $fullName)
{
    global $conf;
    $with_cam_number = true;
    $maxlen = 23;

    if (isset($conf['camname-opts']) && is_array($conf['camname-opts'])) {
        if (array_key_exists('prepend-number', $conf['camname-opts'])) {
            $with_cam_number = (boolean)$conf['camname-opts']['prepend-number'];
        }
        if (array_key_exists('maxlen', $conf['camname-opts'])) {
            $maxlen = (int)$conf['camname-opts']['maxlen'];
        }
    }

    $shortName = '';
    $prefix = '';
    if ($with_cam_number) {
        $prefix .= '(' . $camNr . ') ';
    }
    $prefixLen = strlen($prefix);
    $shortName .= $prefix;

    if (($prefixLen + mb_strlen($fullName)) > $maxlen) {
        $shortName .= mb_substr($fullName, 0, $maxlen - $prefixLen - 3);
        $shortName .= '...';
    } else {
        $shortName .= $fullName;
    }
    return $shortName;
}

/**
 *
 * Функция выводит ссылку нажатия назад в истории браузера
 */
function print_go_back()
{
    print '<br><center><a href="javascript:window.history.back();" title="' . $GLOBALS['strBack'] . '">' .
        '<img src="' . $GLOBALS['conf']['prefix'] . '/img/undo_dark.gif" alt="' . $GLOBALS['strBack'] .
        '" width="24" hspace="24" border="0"></a></center>' . "\n";
}

/**
 *
 * Функция преобразовывает переменную в строку для sql запроса
 * @param  $val
 * @return string
 */
function sql_format_str_val($val)
{
    return empty($val) ? 'NULL' : "'" . addslashes($val) . "'";
}

/**
 *
 * Функция преобразовывает целое число в строку для sql запроса
 * @param  $val
 * @return string
 */
function sql_format_int_val($val)
{
    return empty($val) ? 'NULL' : "'" . (Int)$val . "'";
}

/**
 *
 * Функция преобразовывает число в строку для sql запроса
 * @param  $val
 * @return string
 */
function sql_format_float_val($val)
{
    return empty($val) ? 'NULL' : "'" . str_replace(',', '.', $val) . "'";
}

/**
 *
 * Пустая переменная или нет? (int)0 - не пустая.
 * @param  $var
 * @return boolean
 */
function is_empty_var($var)
{
    return ($var === null || trim((string)$var) === '');
}

function is_non_empty_var($var)
{
    return ($var !== null && trim((string)$var) !== '');
}

/**
 *
 * Функция выводящая ошибку
 * @param string $errstr строка ошибки
 * @param string $file файл
 * @param string $line строка
 */
function MYDIE($errstr = 'internal error', $file = '', $line = '')
{
    if (empty($file)) {
        $file = basename($_SERVER['SCRIPT_FILENAME']);
    }
    print '<div class="error">' . "\n";
    print $errstr;
    print '</div>' . "\n";
    if (!empty($file)) {
        printf('<br><br><div>Error in %s:%d</div>' . "\n", $file, $line);
    }
    if (!isset($GLOBALS['NOBODY'])) {
        print '</body>' . "\n";
    }
    print '</html>' . "\n";
    exit(1);
}

/**
 *
 * Функция записывает в системный лог собщение
 * @param unknown_type $priority Приоритет
 * @param string $message Сообщение
 *
 */
function print_syslog($priority, $message)
{
    if (!isset($GLOBALS['syslog_opened'])) {
        $GLOBALS['syslog_opened'] = 1;
        openlog(
            isset($GLOBALS['AVREG_PROFILE']) ? 'avreg-site-' . $GLOBALS['AVREG_PROFILE'] : 'avreg-site',
            0,
            LOG_DAEMON
        );
    }
    if ($GLOBALS['login_user'] === 'unknown') {
        if (isset($_SERVER['PHP_AUTH_USER'])) {
            $luser = $_SERVER['PHP_AUTH_USER'];
        } elseif (isset($_SERVER['REMOTE_USER'])) {
            $luser = $_SERVER['REMOTE_USER'];
        } else {
            $luser = 'unknown';
        }
    } else {
        $luser = $GLOBALS['login_user'];
    }
    if ($GLOBALS['remote_addr'] === 'unknown') {
        $raddr = $_SERVER['REMOTE_ADDR'];
    } else {
        $raddr = $GLOBALS['remote_addr'];
    }
    syslog($priority, sprintf('%s@%s: %s', $luser, $raddr, $message));
}

/**
 *
 * Функция доступа
 * @param int $good_status статус
 * @param int $http_status http статус
 */
function DENY($good_status = null, $http_status = 403, $do_timeout = false)
{
    global $conf;

    @ob_end_clean();

    /* это исп. для контроля к тем или иным ресурсам в завис. от AVReg-группы  пользователя */
    $user_status = $GLOBALS['user_status'];
    if (!is_null($good_status)) {
        if (settype($good_status, 'int')) {
            if ($user_status <= $good_status) {
                return;
            }
        }
    }

    if ($do_timeout) {
        $tot = @is_int($conf['auth-timeout']) ? $conf['auth-timeout'] : 1;
        sleep(1);
    }

    $aname = & $conf['admin-name'];
    $amail = & $conf['admin-mail'];
    $atel = & $conf['admin-tel'];
    if (!@empty($GLOBALS['grp_ar']) && array_key_exists($user_status, $GLOBALS['grp_ar'])) {
        $deny_reason = sprintf(
            $GLOBALS['access_denided'],
            $GLOBALS['grp_ar'][$user_status]['grname']
        );
        $action = & $GLOBALS['fmtServerAdmin'];
    } else {
        $deny_reason = sprintf(
            @empty($GLOBALS['fmtAccessDenied']) ? "Access denided to &#171;%s@%s&#187" : $GLOBALS['fmtAccessDenied'],
            htmlentities(
                isset($_SERVER['PHP_AUTH_USER']) ? $_SERVER['PHP_AUTH_USER'] : "unknown user",
                ENT_QUOTES,
                $GLOBALS['chset']
            ),
            $_SERVER['REMOTE_ADDR']
        );
        $action = & $GLOBALS['fmtTryOnceMore'];
    }

    if (!empty($aname) || !empty($amail) || !empty($atel)) {
        $deny_reason .= '<br /><br />' . "\n";
        $deny_reason .= sprintf(
            $action,
            '&#034;' . htmlentities($aname, ENT_QUOTES, $GLOBALS['chset']) . '&#034; &#060;' . htmlentities(
                $amail,
                ENT_QUOTES,
                $GLOBALS['chset']
            ) . '&#062; , tel: ' . htmlentities($atel, ENT_QUOTES, $GLOBALS['chset'])
        );
    }

    if ($http_status === 401) {
        if (!isset($_SESSION)) {
            session_start();
        }
        $login_fails_nbr = @intval($_SESSION['login-fails-nbr']);
        $login_fails_nbr += 1;

        $max_401 = (isset($conf['max-401-per-session']) ? max(@intval($conf['max-401-per-session']), 1) : 3);
        if ($login_fails_nbr > $max_401) {
            $http_status = 403;
        } else {
            $_SESSION['login-fails-nbr'] = $login_fails_nbr;
        }
    }

    if (!headers_sent()) {
        switch ($http_status) {
            case 401:
                header('WWW-Authenticate: Basic realm="AVReg server"', true, 401);
                exit();
            case 403:
            default:
                header('Content-Type: text/html; charset=' .$GLOBALS['chset']);
                header('Connection: close', true, 403);
        }
    }

    print_syslog(LOG_CRIT, 'access denided: ' . basename($_SERVER['SCRIPT_FILENAME']));
    print '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">' . "\n";
    print '<html><head>' . "\n";
    print '<link rel="SHORTCUT ICON" href="' . $conf['prefix'] . '/favicon.ico">' . "\n";
    print '<title>403 error::Access Denided</title>' . "\n";
    print '<meta http-equiv="Content-Type" content="text/html; charset=' . $GLOBALS['chset'] . '">' . "\n";
    print "</head><body>\n";
    print '<img src="' . $conf['prefix'] . '/img/password.gif"
        width="48" height="48" border="0"><p><font color="red" size=+1>' . $deny_reason . '</font></p>' . "\n";
    print "</body></html>\r\n";
    exit();
}

/**
 *
 * функция проверяющая входит ли число в промежуток
 * @param unknown_type $int чило
 * @param unknown_type $min минимальное число
 * @param unknown_type $max максимальное число
 * @return bool
 */
function checkIntRange($int, $min, $max)
{
    if ($int == null or $h < $min or $h > $max) {
        return false;
    } else {
        return true;
    }
}

/**
 * Функция возвращает текстовое представление месяца
 *
 * @param int $m месяц от 1 до 12
 * @return string
 */
function getMonth($m)
{
    if (empty($m) or (($m < 0) or ($m > 12))) {
        $retval = $GLOBALS['charNull'];
    } else {
        $retval = $GLOBALS['month_array'][$m - 1];
    }
    return $retval;
}

/**
 * Функция возвращает текстовое представление дня
 *
 * @param int $d день от 1 до 31
 * @return string
 */
function getDay($d)
{
    if (empty($d) or (($d < 1) or ($d > 31))) {
        $retval = $GLOBALS['charNull'];
    } else {
        $retval = sprintf('%02u', $d);
    }
    return $retval;
}

/**
 * Функция возвращает текстовое представление часа
 *
 * @param int $h час от 1 до 23
 * @return string
 */
function getHour($h)
{
    if ($h == null or $h < 0 or $h > 23) {
        $retval = $GLOBALS['charNull'];
    } else {
        $retval = sprintf('%02u', $h);
    }
    return $retval;
}

/**
 * Функция возвращает текстовое представление минут
 *
 * @param int $_min минуты от 1 до 59
 * @return string
 */
function getMinute($_min)
{
    if ($_min == null or $_min < 0 or $_min > 59) {
        $retval = $GLOBALS['charNull'];
    } else {
        $retval = sprintf('%02u', $_min);
    }
    return $retval;
}

/**
 * Функция возвращает текстовое представление дня недели
 *
 * @param string $_weekday перечисление дней в неделе
 * @return string
 */
function getWeekday($_weekday)
{
    $retval = '';
    if ($_weekday == '') {
        $retval = $GLOBALS['charNull'];
    } else {
        $pieces = explode(",", $_weekday);
        foreach ($pieces as $_day) {
            $rrr[] = $GLOBALS['day_of_week'][$_day];
        }
        $retval = implode(",", $rrr);
    }
    return $retval;
}

/**
 * Функция возвращает значение параметра камеры или $retIfEmpty в случ. если оно не задано
 *
 * @param array $camParams - массив с параметрами
 * @param string $camParamName - название параметра
 * @param mixed $retIfEmpty - возвращаемая значение если параметр не найден или его знач. не задано
 * @return mixed
 */
function getCamParamValue($camParams, $camParamName, $retIfEmpty = null)
{
    if (array_key_exists($camParamName, $camParams)) {
        $v = $camParams[$camParamName]['v'];
        if (!is_null($v) && $v !== '') {
            return $v;
        }
    }
    return $retIfEmpty;
}

/**
 *
 * Функция
 * @param string $bin_str
 * @return string
 */
function getBinString($bin_str)
{
    if (strlen($bin_str) == 0) {
        return $bin_str;
    }
    $ret = str_replace("\x00", '\0', $bin_str);
    $ret = str_replace("\x08", '\b', $ret);
    $ret = str_replace("\x0a", '\n', $ret);
    $ret = str_replace("\x0d", '\r', $ret);
    $ret = str_replace("\x1a", '\Z', $ret);
    return $ret;
}

/**
 *
 * Функция проверяющая доступ по паролю
 * @param string $saved_pw сохраненный пароль
 * @param string $pw проверяющий пароль
 */
function check_passwd($saved_pw, $pw)
{
    if ($saved_pw == '' && $pw == '') {
        return;
    }

    $z = crypt($saved_pw, $pw);
    if ($z !== $pw) {
        DENY(null, 401, true);
    }
}

if (!empty($logout)) {
    // print_syslog(LOG_CRIT, '_COOKIE[avreg_logout]: [' . $_COOKIE['avreg_logout'] . ']');
    if (isset($_SESSION)) {
        $_SESSION = array();
    }
    if (isset($_COOKIE) && empty($_COOKIE['avreg_logout'])) {
        // ob_start();
        header('WWW-Authenticate: Basic realm="AVReg server"', true, 401);
        setcookie('avreg_logout', 1); // FIXME google chrome не воспринимает
    } else {
        if ($_SERVER['SERVER_PORT'] == 80) {
            $port_str = '';
        } elseif (empty($_SERVER['SERVER_PORT'])) {
            $port_str = $SSL_ACCESS_USED ? ':443' : ':80';
        } else {
            $port_str = ':' . $_SERVER['SERVER_PORT'];
        }
        header(
            sprintf(
                'Location: %s://%s%s%s%s',
                $SSL_ACCESS_USED ? 'https' : 'http',
                $_SERVER['SERVER_NAME'],
                $port_str,
                $conf['prefix'],
                '/index.php'
            )
        );
        setcookie('avreg_logout', 0);
    }
    setcookie('as_guest', 0, -100);
    unset($_COOKIE['as_guest']);

    exit();
}

/*********************** Код: Читаем конфиг. файл(ы) и др. ******************************/
try {
    $res = confparse($conf, 'avreg-site');
} catch (ErrorException $e) {
    $code = $e->getCode();
    if ( $code === 13 /* EACCES */ ) {
        DENY();
    } else {
        throw($e);
    }
} catch (Exception $e) {
    throw($e);
}

if ($conf['prefix'] !== '/') {
    $cp = session_get_cookie_params();
    // FIXME WTF "lifetime"?
    session_set_cookie_params($cp['lifetime'], $conf['prefix'] . '/');
    unset($cp);
}

// XXX session need to external http://login:pass@server/avreg/ access
if (!@empty($USE_SESSION)) {
    if (is_string($USE_SESSION)) {
        session_name($USE_SESSION);
    }
    session_start();
}

function array_merge_recursive_unique($array1, $array2) {
  if (empty($array1)) return $array2; //optimize the base case

  foreach ($array2 as $key => $value) {
    if (is_array($value) && is_array(@$array1[$key])) {
      $value = array_merge_recursive_unique($array1[$key], $value);
    }
    $array1[$key] = $value;
  }
  return $array1;
}

$conf = array_merge_recursive_unique($conf, $res);

$EXISTS_PROFILES = array();
unset($AVREG_PROFILE);
if (!empty($conf['prefix']) && preg_match('@^/([^/]+).*@', $_SERVER['REQUEST_URI'], $matches)) {
    if (strcasecmp($matches[1], 'avreg') === 0) {
        $EXISTS_PROFILES = get_avreg_profiles($conf);
    } else {
        $res = confparse($conf, 'avreg-site', $conf['profiles-dir'] . '/' . $matches[1]);
        if (!$res) {
            die("<br /><br />Error: not found active profile " . $conf['profiles-dir'] . '/' . $matches[1]);
        }
        $AVREG_PROFILE = $matches[1];
        // tohtml($res);
        if (is_array($res)) {
            $conf = array_merge_recursive_unique($conf, $res);
            $conf['prefix'] = '/' . $AVREG_PROFILE;
            $conf['daemon-name'] .= '-' . $AVREG_PROFILE;
        }
    }
} else {
    $EXISTS_PROFILES = get_avreg_profiles($conf);
}

/// XXX after confparse
setlocale(LC_ALL, $locale_str);
///  путь к языковым файлам
$lang_dir = $wwwdir . 'lang/' . $lang . '/' . strtolower($chset) . '/';
///  языковый модуль
$lang_module_name = $lang_dir . 'common.inc.php';
///  название локального языкового модуля
$pt = substr($_SERVER['PHP_SELF'], strlen($conf['prefix']));
///  параметры языкового модуля
$params_module_name = $lang_dir . 'params.inc.php';
require($lang_module_name);
///  языковый модуль
$lang_module_name2 = $lang_dir . str_replace('/', '_', $pt);
if (isset($lang_file)) {
    $lang_module_name2 = $lang_dir . $lang_file;
}
if (file_exists($lang_module_name2)) {
    require($lang_module_name2);
}

if ($conf['debug']) {
    ini_set('display_errors', '1');
    ini_set('log_errors', '1');
    ini_set('html_errors', '1');
    error_reporting(
        E_ERROR | E_WARNING | E_PARSE | E_CORE_ERROR | E_CORE_WARNING |
        E_COMPILE_ERROR | E_COMPILE_WARNING | E_USER_ERROR | E_USER_WARNING |
        E_USER_NOTICE
    );
}

require_once('adb.php');
$result = $adb->getUsers();

foreach ($result as $row) {

    /* var_dump($row); */
    $ui = array();
    $ui['HOST'] = $row['HOST'];
    $ui['USER'] = $row['USER'];
    $ui['PASSWD'] = $row['PASSWD'];
    $ui['STATUS'] = (int)$row['STATUS'];
    $ui['GUEST'] = (int)$row['GUEST'];
    $ui['PDA'] = (int)$row['PDA'];
    $ui['ALLOW_CAMS'] = $row['ALLOW_CAMS'];
    $ui['ALLOW_LAYOUTS'] = $row['ALLOW_LAYOUTS'];
    $ui['MAX_FORCED_REC_MINUTES'] = is_null(
        $row['MAX_FORCED_REC_MINUTES']
    ) ? null : (int)$row['MAX_FORCED_REC_MINUTES'];
    $ui['MAX_MEDIA_SESSIONS_NB'] = is_null($row['MAX_MEDIA_SESSIONS_NB']) ? null : (int)$row['MAX_MEDIA_SESSIONS_NB'];
    $ui['MAX_VIDEO_FPS'] = is_null($row['MAX_VIDEO_FPS']) ? null : $row['MAX_VIDEO_FPS'];
    $ui['MAX_VIDEO_NONMOTION_FPS'] = is_null($row['MAX_VIDEO_NONMOTION_FPS']) ? null : $row['MAX_VIDEO_NONMOTION_FPS'];
    $ui['MAX_MEDIA_SESSION_RATE_KB'] = is_null(
        $row['MAX_MEDIA_SESSION_RATE_KB']
    ) ? null : (int)$row['MAX_MEDIA_SESSION_RATE_KB'];
    $ui['MAX_MEDIA_SESSION_MINUTES'] = is_null(
        $row['MAX_MEDIA_SESSION_MINUTES']
    ) ? null : (int)$row['MAX_MEDIA_SESSION_MINUTES'];
    $ui['MAX_MEDIA_SESSION_VOLUME_MB'] = is_null(
        $row['MAX_MEDIA_SESSION_VOLUME_MB']
    ) ? null : (int)$row['MAX_MEDIA_SESSION_VOLUME_MB'];
    $ui['LONGNAME'] = $row['LONGNAME'];
    $ui['CHANGE_HOST'] = $row['CHANGE_HOST'];
    $ui['CHANGE_USER'] = $row['CHANGE_USER'];
    $ui['CHANGE_TIME'] = $row['CHANGE_TIME'];
    $users[] = $ui;
}
unset($result);

//режим гостя
$GuestAuth = false;
$as_guest = false;
if (isset($_GET['user'])) {
    $as_guest = trim($_GET['user']);
    setcookie('as_guest', $as_guest);
}
if (isset($_COOKIE['as_guest']) && !$as_guest) {
    $as_guest = trim($_COOKIE['as_guest']);
}
if (isset($as_guest)) {

    foreach ($GLOBALS['users'] as $key => $val) {
        if ($val["USER"] == $as_guest && $val["GUEST"] == 1) {

            $GuestAuth = true;

            $_SERVER['REMOTE_USER'] = $_SERVER['PHP_AUTH_USER'] = $as_guest;
            break;
        }
    }
}

if ( $GuestAuth ) {
    $_SERVER['PHP_AUTH_PW'] = '';
}

/// Альтернативная внешняя аутентификация.
$ExternalAuth = false;

if (!empty($_SERVER['AUTH_TYPE']) && !empty($_SERVER['REMOTE_USER'])
    && !empty($conf['ExternalAuthMap'])
    && file_exists($conf['ExternalAuthMap']) && !$GuestAuth
) {
    try {
        $cfm = check_file_mode($conf['ExternalAuthMap'], 0, 999, 6,4,0);
        if (!$cfm) {
            throw new InvalidArgumentException ("ExternalAuthMap file \"$conf[ExternalAuthMap]\" should have a \"root:www-data 0640\" perms");
            clearstatcache(); // to next request
        }
    } catch (ErrorException $e) {
        // stat() failed
        clearstatcache(); // non need if ENOENT but ...
        throw($e);
    }

    $lines = file($conf['ExternalAuthMap']);
    if (false === $lines || count($lines) < 1) {
        throw new InvalidArgumentException ("could not read the ExternalAuthMap \"$conf[ExternalAuthMap]\" file");
    }
    foreach ($lines as &$line) {
        $l = trim($line);
        if (empty($l) || preg_match('/^[#;]/' , $l)) {
            continue;
        }
        @list($ruser, $auser_info) = explode('=', trim($line));
        if (!empty($auser_info)) {
            if (trim($ruser) === $_SERVER['REMOTE_USER']) {
                $ExternalAuth = true;
                break;
            }
        }
    }
    unset($lines, $line, $l, $ruser, $luser);
    if ($ExternalAuth) {
        $a = explode(':', trim($auser_info), 2);
        $_SERVER['REMOTE_USER'] = trim($a[0]);
        if (isset($a[1])) {
            $_SERVER['PHP_AUTH_PW'] = trim($a[1]);
        }
    } else {
        DENY(null, 403, true);
    }
    unset($auser_info, $a);
}

if (isset($_SERVER['PHP_AUTH_USER'])) {
    require_once($wwwdir . 'lib/utils-inet.php');
    if (!isset($_SERVER['REMOTE_USER'])) {
        $_SERVER['REMOTE_USER'] = $_SERVER['PHP_AUTH_USER'];
    }

    $user_info = avreg_find_user(ip2long($_SERVER['REMOTE_ADDR']), -1, $_SERVER['REMOTE_USER']);
    if ($user_info === false) {
        DENY(null, 401, true);
    }

    if (!$GuestAuth) {
        check_passwd($_SERVER['PHP_AUTH_PW'], $user_info['PASSWD']);
    }

    $login_user = & $_SERVER['REMOTE_USER'];
    $user_status = & $user_info['STATUS'];
    $login_user_name = & $row['LONGNAME'];
    $login_host = & $remote_addr;
    $ALLOW_CAMS = parse_dev_acl($user_info['ALLOW_CAMS']);
    if ($user_status <= $install_status && !$GuestAuth) {
        $install_user = true;
    }
    if ($user_status <= $admin_status && !$GuestAuth) {
        $admin_user = true;
    }
    if ($user_status <= $arch_status) {
        $arch_user = true;
    }
    if ($user_status <= $operator_status) {
        $operator_user = true;
    }
    if ($user_status <= $viewer_status) {
        $viewer_user = true;
    }

    $allow_pda = $user_info['PDA'];

} else {
    DENY(null, 401, false); // если первый чистый вход - сразу 401 без таймаута
}

if (isset($_SESSION) && isset($_SESSION['login-fails-nbr'])) {
    $_SESSION['login-fails-nbr'] = 1; // to ?logout=1 case
}
