<?php
/**
 * Copyright (C) 2004-2021 "Setevye informacionnye sistemy", Ltd (Russia)
 *                          en: "Network Information Systems"
 */

namespace Avreg;

class Cache
{
    const APP_PREFIX = 'avreg';

    private $memcache;
    private $cache_id = '';
    private $locked = ':lock';

    private $pconnect = true;
    private $server = 'localhost';
    private $port = 11211;
    private $debug = false;

    protected static $defaultOptions = array(
        'prefix' => '',
        'lock-expire' => 30,
        'key-expire'  => 3600,
        'debug' => ''
    );
    private $options;

    public function __construct(array $options = null)
    {
        if (is_null($options)) {
            $options = array();
        }
        $this->options = array_merge(self::$defaultOptions, $options);

        $this->memcache = new \Memcached;
        $this->memcache->addServer($this->server, $this->port);
        // test connection
        $keyName = $this->keyName('test');
        $this->memcache->set($keyName, 5);  /* FIXME race cond ?*/
        $k = $this->memcache->get($keyName);
        if ($k !== 5) {
            throw new \Exception(get_class()  . "::addServer($this->server, $this->port) failed: " . $this->memcache->getResultMessage());
        }

        if ($this->options['debug']) {
            error_log(sprintf('%s(%s, %s)', __METHOD__, $this->options['prefix'], (string)$this->options['debug']));
        }
    }

    public function __destruct()
    {
        $this->memcache->quit();
        if ($this->options['debug']) {
            error_log(sprintf('%s()', $this->options['prefix'], __METHOD__));
        }
    }

    protected function keyName(string $key)
    {
        if (empty($this->options['prefix'])) {
            return self::APP_PREFIX . ':' . $key;
        } else {
            return self::APP_PREFIX . ':' . $this->options['prefix'] . ':' . $key;
        }
    }

    protected function lockName(string $key)
    {
        return $this->keyName($key) . ':lock';
    }

    public function get(string $key)
    {
        $keyName = $this->keyName($key);
        $data = $this->memcache->get($keyName);
        if ($this->options['debug']) {
            error_log(sprintf('%s("%s") -> %s%s', __METHOD__, $keyName, gettype($data), @empty($data) ? '(empty)' : ''));
        }
        return $data;
    }

    private function waitLock(string $key, int $wait = null)
    {
        if (is_null($wait)) {
            $wait = $this->options['lock-expire'];
        }

        $lockName = $this->lockName($key);

        $t_start = microtime(TRUE);
        do {
            if ($this->memcache->add($lockName, 'locked', $this->options['lock-expire'])) {
                break;
            }
            if ($this->memcache->getResultCode() !== \Memcached::RES_NOTSTORED) {
                sleep(1);
            }
            if (microtime(TRUE) - $t_start > $wait) {
                throw new \Exception(
                    sprintf("%s::lock(%s) failed: %u sec. timeout has occurred",  get_class(), $key, $wait)
                );
            }
        } while (true);
    }

    public function lock(string $key, int $wait = null)
    {
        $this->waitLock($key, $wait);

        $keyName = $this->keyName($key);
        $ret = $this->memcache->get($keyName);
        if ($this->memcache->getResultCode() === \Memcached::RES_NOTFOUND) {
            $ret =  array();
        }

        if ($this->options['debug']) {
            error_log(sprintf('%s("%s,") -> %s', __METHOD__, $keyName, print_r($ret, true)));
        }
        return $ret;
    }

    public function unlock(string $key, $data = null)
    {
        $lockName = $this->lockName($key);
        $keyName = $this->keyName($key);

        $ret = true;

        if (!is_null($data)) {
            $ret = $this->memcache->set($keyName, $data, $this->options['key-expire']);
        }

        $this->memcache->delete($lockName);

        if ($this->options['debug']) {
            error_log(sprintf('%s("%s,") -> %s', __METHOD__, $keyName, $ret));
        }
        return $ret;
    }

    public function set(string $key, $data, int $wait = null)
    {
        $this->waitLock($key, $wait);
        return $this->unlock($key, $data);
    }

    public function delete(string $key, int $wait = null)
    {
        $this->waitLock($key, $wait);

        $keyName = $this->keyName($key);
        $ret = $this->memcache->delete($keyName);
        if ($this->options['debug']) {
            error_log(sprintf('%s("%s") -> %s', __METHOD__, $keyName, $ret));
        }

        $this->memcache->delete($this->lockName($keyName));

        return $ret;
    }

    public function flush()
    {
        $ret = $this->memcache->flush();
        if ($this->options['debug']) {
            error_log(sprintf('%s("") -> %s', __METHOD__, $ret));
        }
        return $ret;
    }
}
