<?php
/**
 * Copyright (C) 2004-2021 "Setevye informacionnye sistemy", Ltd (Russia)
 *                          en: "Network Information Systems"
 */

// TODO http_throttle

namespace Avreg\Playlist\Zip;

// @codingStandardsIgnoreStart
class InputFileException extends \Exception
{
    public function __construct($path)
    {
        parent::__construct("File \"$path\" not exist or access denided'");
    }
}

class InputOutputException extends \Exception
{
    public function __construct($path)
    {
        parent::__construct("Error while reading file \"$path\"");
    }
}
class TooLargeZip extends \Exception
{
    public function __construct($zipName, $len, $maxLen)
    {
        parent::__construct("The length ($len) of the $zipName file can not be larger than $maxLen bytes.");
    }
}
// @codingStandardsIgnoreEnd

/**
 * Class to dynamically create a zip file and put to stdout
 */
class ZipSink
{
    /**
     * Entry source types
     */
    const ENTRY_SRC_IS_DIR = 0;
    const ENTRY_SRC_IS_FILE = 1;
    const ENTRY_SRC_IS_RAWDATA = 2;

    const OUTPUT_BLOCK_SZ_DEFAULT = 65536;

    const ZIP_MAX_LEN = 4294967295;

    private $zipLength = false;

    private $filesTree = array();
    private $rootRawMap = array();

    // counters of input elements
    private $dirs = 0;
    private $files = 0;
    private $rawdatas = 0;

    private $entryAllDataLen = 0;
    private $entryAllNamesLen = 0;

    private $centralDirectory = array(); // central directory
    private $offset = 0;

    public function __construct($filesMap, &$rootRawMap = false)
    {
        $this->now = time();
        $this->chunk_sz = self::OUTPUT_BLOCK_SZ_DEFAULT;

        $this->buildFilesTree($filesMap);
        if ($rootRawMap && is_array($rootRawMap)) {
            $this->rootRawMap = $rootRawMap;
        }

        $this->calcLength();

        if ($this->zipLength > self::ZIP_MAX_LEN) {
            throw new TooLargeZip('ZIP', $this->zipLength, self::ZIP_MAX_LEN);
        }
    } /* __construct() */

    public function getLength()
    {
        return $this->zipLength;
    } /* getLength() */

    public function out()
    {
        while (@ob_end_flush()) {
        };

        // files
        $iterator = new \RecursiveIteratorIterator(
            new \RecursiveArrayIterator($this->filesTree),
            \RecursiveIteratorIterator::SELF_FIRST
        );

        foreach ($iterator as $key => $value) {
            if ($iterator->hasChildren()) {
                // dir
                $this->putDir($key);
            } else {
                // file
                $this->putFile($key, $value);
            }
        }

        while (@ob_end_flush()) {
        };

        // root object(s)
        foreach ($this->rootRawMap as &$rawObj) {
            $zipfileName = $rawObj['target'];
            $this->putZipEntry(
                self::ENTRY_SRC_IS_RAWDATA,
                $zipfileName,
                $this->now,
                $rawObj['blob']['data'],
                $rawObj['blob']['length']
            );
        }

        $this->putTrailer();
    } /* out() */

    protected function putDir($zipPathHash)
    {
        $a = explode('|', $zipPathHash);
        $entryPath = $a[0];
        if (isset($a[1])) {
            $entryMtime = (int)$a[1];
        } else {
            $entryMtime = $this->now;
        }
        $this->putZipEntry(
            self::ENTRY_SRC_IS_DIR,
            $entryPath . '/',
            $entryMtime
        );
    } /* putDir() */

    protected function putFile($zipPathHash, $absSrcFilePath)
    {
        $a = explode('|', $zipPathHash);
        $entryPath = $a[0];
        if (isset($a[1])) {
            $entryMtime = (int)$a[1];
        } else {
            $entryMtime = $this->now;
        }
        if (isset($a[1])) {
            $entryMtime = (int)$a[1];
        } else {
            $entryMtime = $this->now;
        }
        if (isset($a[2])) {
            $srcSize = (int)$a[2];
        } else {
            $srcSize = filesize($absSrcFilePath);
        }
        $this->putZipEntry(
            self::ENTRY_SRC_IS_FILE,
            $entryPath,
            $entryMtime,
            $absSrcFilePath,
            $srcSize
        );
    } /* putFile() */

    private function getCompHdr($entryType)
    {
        $ret = '';
        $bitflag = 0;

        if ($entryType === self::ENTRY_SRC_IS_FILE) {
            $bitflag |= 0x08; // crc32, compLen, uncompLen true un data descriptor
        }

        $ret .= "\x1a\x00"; // version to extract
        $ret .= pack('v', $bitflag); // general purpose bit flag
        $ret .= "\x00\x00"; // compression method

        return $ret;
    } /* getCompHdr() */

    private function sendFile($filePath)
    {
        $ctx = hash_init('crc32b');
        $fd = fopen($filePath, 'rb');
        while (!feof($fd)) {
            $chunk = fread($fd, $this->chunk_sz);
            if ($chunk === false) {
                fclose($fd);
                throw new InputOutputException($filePath);

            }
            echo($chunk);
            hash_update($ctx, $chunk);
            // ob_flush();
            flush();
            if (connection_status() != 0) {
                fclose($fd);
                die();
            }
        }
        fclose($fd);
        $crc32Code = hexdec(hash_final($ctx));
        return $crc32Code;
    } /* sendFile() */

    private function putZipEntry($entryType, $entryPath, $entryMtime, $entrySrc = null, $srcLen = 0)
    {
        if ($entryType === self::ENTRY_SRC_IS_DIR || !$entrySrc || $srcLen === 0) {
            $entrySrc = null;
            $srcLen = 0;
        }
        $extFileAttr = $entryType === self::ENTRY_SRC_IS_DIR ? 16 : 32;

        $msDateTime = $this->getMsDosDateTime($entryMtime);
        $entryPathLen = strlen($entryPath);
        $srcCRC32 = 0;
        $compLen = 0;
        $uncompLen = 0;

        $localFileHeader = '';
        $dataDescriptor = '';
        $centralDirStruct = '';

        $compHdr = $this->getCompHdr($entryType);

        if ($srcLen > 0 && $entryType === self::ENTRY_SRC_IS_RAWDATA) {
            $srcCRC32 = crc32($entrySrc);
            $compLen = $uncompLen = $srcLen;
        }

        // 4.3.7 Local file header
        $localFileHeader .= "\x50\x4b\x03\x04";             // signature
        $localFileHeader .= $compHdr;                       // version to extract, bit flags, compression method
        $localFileHeader .= $msDateTime['time'];            // mtime time
        $localFileHeader .= $msDateTime['date'];            // mtime date
        $localFileHeader .= pack('V', $srcCRC32);           // CRC32
        $localFileHeader .= pack('V', $compLen);            // comp. length
        $localFileHeader .= pack('V', $uncompLen);             // uncomp. length
        $localFileHeader .= pack('v', strlen($entryPath));  // file name length
        $localFileHeader .= pack('v', 0);                  // extra field length
        $localFileHeader .= $entryPath;                     // file name (variable size)
        // $localFileHeader .= $extra;                      // extra field (variable size)
        echo($localFileHeader);

        // 4.3.8  File data
        if ($srcLen > 0) {
            if ($entryType === self::ENTRY_SRC_IS_RAWDATA) {
                echo($entrySrc);
            } elseif ($entryType === self::ENTRY_SRC_IS_FILE) {
                $srcCRC32 = $this->sendFile($entrySrc);
                $compLen = $uncompLen = $srcLen;

                // 4.3.9  Data descriptor
                // $dataDescriptor .= "\x50\x4b\x07\x08";
                $dataDescriptor .= pack('V', $srcCRC32);
                $dataDescriptor .= pack('V', $compLen);
                $dataDescriptor .= pack('V', $uncompLen);
                echo($dataDescriptor);
            }
        }

        // 4.3.12  Central directory structure:
        $centralDirStruct .=  "\x50\x4b\x01\x02";  // central file header signature
        $centralDirStruct .= "\x00\x00";          // version made by
        $centralDirStruct .= $compHdr;            // version to extract, bit flags, compression method
        $centralDirStruct .= $msDateTime['time']; // mtime time
        $centralDirStruct .= $msDateTime['date']; // mtime date
        $centralDirStruct .= pack('V', $srcCRC32);
        $centralDirStruct .= pack('V', $compLen);
        $centralDirStruct .= pack('V', $uncompLen);
        $centralDirStruct .= pack('v', strlen($entryPath)); // file name length
        $centralDirStruct .= pack('v', 0);        // extra field length
        $centralDirStruct .= pack('v', 0);        // file comment length
        $centralDirStruct .= pack('v', 0);        // disk number start
        $centralDirStruct .= pack('v', 0);        // internal file attributes
        $centralDirStruct .= pack('V', $extFileAttr); // external file attributes
        $centralDirStruct .= pack('V', $this->offset); // relative offset of local header 4 bytes
        // file name (variable size)
        // extra field (variable size)
        // file comment (variable size)
        $centralDirStruct .= $entryPath;

        $this->offset += strlen($localFileHeader) + $compLen + strlen($dataDescriptor);
        $this->centralDirectory[] = $centralDirStruct;
    } /* addFile() */

    private function putTrailer()
    {
        $controlDirectory = implode("", $this->centralDirectory);
        $controlDirectoryLen = strlen($controlDirectory);
        $centralDirectoryNbrs = count($this->centralDirectory);

        // central directory record
        echo($controlDirectory);

        echo "\x50\x4b\x05\x06". // end of central directory record
             "\x00\x00". // number of this disk
             "\x00\x00". // number of the disk with the start of the central directory
             pack('v', $centralDirectoryNbrs). // Number of central directory records on this disk
             pack('v', $centralDirectoryNbrs). // Total number of central directory records
             pack('V', $controlDirectoryLen).  // Size of central directory (bytes)
             pack('V', $this->offset). // Offset of start of central directory, relative to start of archive
             "\x00\x00"; // Comment length (n)
    } /* putTrailer() */

    protected function calcLength()
    {
        // calc root small obj
        //
        foreach ($this->rootRawMap as &$rawObj) {
            $this->rawdatas++;
            $this->entryAllDataLen += $rawObj['blob']['length'];
            $this->entryAllNamesLen += strlen($rawObj['target']);
        }

        // calc files
        $iterator = new \RecursiveIteratorIterator(
            new \RecursiveArrayIterator($this->filesTree),
            \RecursiveIteratorIterator::SELF_FIRST
        );

        foreach ($iterator as $key => $value) {
            $ar = explode('|', $key);
            $entryPath = $ar[0];
            $this->entryAllNamesLen += strlen($entryPath);
            if ($iterator->hasChildren()) {
                // dir
                $this->dirs++;
                $this->entryAllNamesLen += 1; // trailing / in putDir
            } else {
                // file
                $this->files++;
                $this->entryAllDataLen += (int)$ar[2];
            }
        }

        $LFH_STATIC_SZ = 30;
        $DD_SZ = 12;
        $CENTDIRH_STATIC_SZ = 46;
        $EOCD_STATIC_SZ = 22;

        $d = $this->dirs;
        $f = $this->files;
        $o = $this->rawdatas;
        $a = $d + $f + $o;

        $this->zipLength = $a * ($LFH_STATIC_SZ + $CENTDIRH_STATIC_SZ) +
            $f * $DD_SZ +
            $this->entryAllNamesLen * 2 +
            $this->entryAllDataLen +
            $EOCD_STATIC_SZ;
    } /* calcLength() */

    protected function buildFilesTree($filesMap)
    {
        $a = array();
        foreach ($filesMap as &$v) {
            $filePath = $this->normalizePath($v['src']);
            $target_chunks = explode('/', $v['target']);
            $deep_max = count($target_chunks);
            $target_tree = array();
            $tlink =& $target_tree;
            $deep = 1;
            foreach ($target_chunks as &$c) {
                $pathPrefix = '';
                if ($deep > 1) {
                    $pathPrefix = implode('/', array_slice($target_chunks, 0, $deep - 1)) . '/';
                }
                if ($deep < $deep_max) {
                    // directory
                    $key = $pathPrefix . $c . '|' . $this->now; // TODO get real mtime if available
                    $tlink[$key] = array();
                    $tlink =& $tlink[$key];
                } else {
                    // file
                    $stat = @stat($filePath);
                    if ($stat === false) {
                        throw new InputFileException($v['src']);
                    }
                    $key = $pathPrefix . $c . '|' . $stat['mtime'] . '|' . $stat['size'];
                    $tlink[$key] = $filePath;
                }
                $deep++;
            }
            $a = array_merge_recursive($a, $target_tree);
        }
        $this->filesTree = $a;
        // file_put_contents("/tmp/filesTree.txt", print_r($this->filesTree, true));
    } /* buildFilesTree() */

    public function getMsDosDateTime($unixtime)
    {
        $unixtime = ($unixtime + 1) & (~1);
        $lt = localtime($unixtime, true);
        $sec = $lt['tm_sec'];
        $min = $lt['tm_min'];
        $hour = $lt['tm_hour'];
        $ms_time = 0xFFFF & ($hour << 11 | $min << 5 | $lt >> 1);

        $day = $lt['tm_mday'];
        $month = ($lt['tm_mon'] + 1);
        $lt['tm_year'] += 1900;
        if ($lt['tm_year'] < 1980) {
            $year = 1080;
        } else {
            $year = $lt['tm_year'];
        }
        $year -= 1980;
        $ms_date = 0xFFFF & ($year << 9 | $month << 5 | $day);

        return array(
            'date' => pack('v', $ms_date),
            'time' => pack('v', $ms_time)
        );
    } /* getMsDosDateTime() */

    public function normalizePath($path)
    {
        $path = str_replace('\\', '/', $path); // Replace backslashes with forwardslashes
        $path = preg_replace('/\/+/', '/', $path); // Combine multiple slashes into a single slash
        $path = preg_replace('/\/$/', '', $path); // Remove trailing /
        return $path;
    } /* normalizePath() */
}

/* vim: set expandtab smartindent tabstop=4 shiftwidth=4: */
